import pytest
import torch
from botorch.posteriors.gpytorch import GPyTorchPosterior

from axtreme.sampling import UTSampler
from tests.sampling.helpers import (
    posterior_b2_n3_t2,
    posterior_n1_t1,
    posterior_n1_t2,
    posterior_n2_t1,
    posterior_n3_t2,
)


@pytest.mark.parametrize(
    "posterior, expected_base_sample_shape",
    [
        (posterior_n1_t1(), torch.Size([3, 1])),  # MultivariateNormal
        (posterior_n1_t2(), torch.Size([5, 2])),  # MultitaskMultivariateNormal
    ],
)
def test_construct_base_samples_created_with_right_shape(
    posterior: GPyTorchPosterior, expected_base_sample_shape: torch.Size
):
    """NOTE: the sample_shape for UT is determined by the size of the input.

    sample_shape = 2 * t + 1, where t comes from the posterior shape (*b,n,t)
    """
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)

    assert sampler.base_samples.shape == expected_base_sample_shape


@pytest.mark.parametrize(
    "posterior1, posterior2",
    [
        (posterior_n1_t1(), posterior_n2_t1()),  # MultivariateNormal
        (posterior_n1_t2(), posterior_n3_t2()),  # MultitaskMultivariateNormal
        (posterior_n1_t2(), posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
    ],
)
def test_construct_base_samples_reuse_base_samples_when_possible(
    posterior1: GPyTorchPosterior, posterior2: GPyTorchPosterior
):
    """Reuse base samples if the samples shape and output dim t don't change."""
    sampler = UTSampler()
    sampler._construct_base_samples(posterior1)
    original_base_samples = id(sampler.base_samples)

    sampler._construct_base_samples(posterior2)
    base_samples_after_repeat = id(sampler.base_samples)

    # The same obect should be used
    assert original_base_samples == base_samples_after_repeat


def test_construct_base_samples_for_posterior_with_new_target_size():
    sampler = UTSampler()
    sampler._construct_base_samples(posterior_n1_t1())
    base_samples_1 = sampler.base_samples
    sampler._construct_base_samples(posterior_n1_t2())
    base_samples_2 = sampler.base_samples

    assert base_samples_1.shape == torch.Size([3, 1])
    assert base_samples_2.shape == torch.Size([5, 2])


@pytest.mark.parametrize(
    "posterior, expected_sample_shape",
    [
        (posterior_n1_t1(), torch.Size([3, 1, 1])),  # MultivariateNormal
        (posterior_n1_t2(), torch.Size([5, 1, 2])),  # MultitaskMultivariateNormal
        (posterior_n3_t2(), torch.Size([5, 3, 2])),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), torch.Size([5, 2, 3, 2])),  # MultitaskMultivariateNormal batched
    ],
)
def test_forward(posterior: GPyTorchPosterior, expected_sample_shape: torch.Size):
    sampler = UTSampler()
    samples = sampler(posterior)

    assert samples.shape == expected_sample_shape


@pytest.mark.parametrize(
    "posterior, sample_shape, expected_mean_shape",
    [
        (posterior_n1_t1(), (3, 1), (1,)),  # MultivariateNormal
        (posterior_n1_t1(), (3, 2, 2), (2, 2)),  # MultivariateNormal
        (posterior_n1_t2(), (5, 3), (3,)),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), (5, 4), (4,)),  # MultitaskMultivariateNormal batched
    ],
)
def test_mean_shape(posterior: GPyTorchPosterior, sample_shape: torch.Size, expected_mean_shape: torch.Size):
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)
    samples = torch.randn(sample_shape)
    mean = sampler.mean(samples, dim=0)

    assert mean.shape == expected_mean_shape


@pytest.mark.parametrize(
    "posterior, dim, sample_shape, expected_mean_shape",
    [
        (posterior_n1_t1(), 0, (3, 1), (1,)),  # MultivariateNormal
        (posterior_n1_t1(), 1, (2, 3, 2), (2, 2)),  # MultivariateNormal
        (posterior_n1_t2(), 1, (3, 5), (3,)),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), 2, (1, 1, 5), (1, 1)),  # MultitaskMultivariateNormal
    ],
)
def test_mean_different_dim(
    posterior: GPyTorchPosterior,
    dim: int,
    sample_shape: torch.Size,
    expected_mean_shape: torch.Size,
):
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)
    samples = torch.randn(sample_shape)
    mean = sampler.mean(samples, dim=dim)

    assert mean.shape == expected_mean_shape


@pytest.mark.parametrize(
    "posterior, dim, sample_shape",
    [
        (posterior_n1_t1(), 0, (1, 1)),  # MultivariateNormal
        (posterior_n1_t1(), 1, (3, 1)),  # MultivariateNormal
        (posterior_n1_t2(), 0, (3, 5)),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), 1, (5, 4)),  # MultitaskMultivariateNormal batched
    ],
)
def test_mean_fail_when_wrong_shape(
    posterior: GPyTorchPosterior,
    dim: int,
    sample_shape: torch.Size,
):
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)
    samples = torch.randn(sample_shape)
    with pytest.raises(ValueError):
        _ = sampler.mean(samples, dim=dim)


@pytest.mark.parametrize(
    "posterior",
    [
        (posterior_n1_t1()),  # MultivariateNormal
        (posterior_n1_t2()),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
    ],
)
def test_mean_value(posterior: GPyTorchPosterior):
    sampler = UTSampler()
    samples = sampler(posterior)
    mean = sampler.mean(samples)
    posterior_mean = posterior.mean

    assert torch.allclose(mean, posterior_mean, atol=1e-10)


@pytest.mark.parametrize(
    "posterior, sample_shape, expected_var_shape",
    [
        (posterior_n1_t1(), (3, 1), (1,)),  # MultivariateNormal
        (posterior_n1_t1(), (3, 2, 2), (2, 2)),  # MultivariateNormal
        (posterior_n1_t2(), (5, 3), (3,)),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), (5, 4), (4,)),  # MultitaskMultivariateNormal batched
    ],
)
def test_var_shape(
    posterior: GPyTorchPosterior,
    sample_shape: torch.Size,
    expected_var_shape: torch.Size,
):
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)
    samples = torch.randn(sample_shape)
    var = sampler.var(samples, dim=0)

    assert var.shape == expected_var_shape


@pytest.mark.parametrize(
    "posterior, dim, sample_shape, expected_var_shape",
    [
        (posterior_n1_t1(), 0, (3, 1), (1,)),  # MultivariateNormal
        (posterior_n1_t1(), 1, (2, 3, 2), (2, 2)),  # MultivariateNormal
        (posterior_n1_t2(), 1, (3, 5), (3,)),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), 2, (1, 1, 5), (1, 1)),  # MultitaskMultivariateNormal batched
    ],
)
def test_var_different_dim(
    posterior: GPyTorchPosterior,
    dim: int,
    sample_shape: torch.Size,
    expected_var_shape: torch.Size,
):
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)
    samples = torch.randn(sample_shape)
    var = sampler.var(samples, dim=dim)

    assert var.shape == expected_var_shape


@pytest.mark.parametrize(
    "posterior, dim, sample_shape",
    [
        (posterior_n1_t1(), 0, (1, 1)),  # MultivariateNormal
        (posterior_n1_t1(), 1, (3, 1)),  # MultivariateNormal
        (posterior_n1_t2(), 0, (3, 5)),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), 1, (5, 4)),  # MultitaskMultivariateNormal batched
    ],
)
def test_var_fail_when_wrong_shape(
    posterior: GPyTorchPosterior,
    dim: int,
    sample_shape: torch.Size,
):
    sampler = UTSampler()
    sampler._construct_base_samples(posterior)
    samples = torch.randn(sample_shape)
    with pytest.raises(ValueError):
        _ = sampler.var(samples, dim=dim)


@pytest.mark.parametrize(
    "posterior",
    [
        (posterior_n1_t1()),  # MultivariateNormal
        (posterior_n1_t2()),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
    ],
)
def test_var_value(posterior: GPyTorchPosterior):
    sampler = UTSampler()
    samples = sampler(posterior)
    var = sampler.var(samples)
    posterior_var = posterior.variance

    assert torch.allclose(var, posterior_var, atol=1e-10)



# What are the key things to check?

# only support 2d
# that the order doesnt change with different batch sizes

import pytest
import torch

from axtreme.data.batch_invariant_sampler import BatchInvariantSampler2d


@pytest.mark.parametrize(
    "batch_shape",
    [
        # 2d check
        (torch.Size([2, 5])),  # partial batch of size torch.Size([2,1])
    ],
)
def test_batch_invariant_sampler_partial_batchs(batch_shape: torch.Size):
    # Typically this will be a Sampler object, but an iterator is perfectly valid, and is simpler to is used here
    # sourcery skip: no-loop-in-tests
    samples = list(range(12))
    batch_sampler = BatchInvariantSampler2d(samples, batch_shape=batch_shape)

    # Check the standard batches
    batches = list(batch_sampler)
    for batch in batches[:-1]:
        # print this if want more viual debugging
        assert torch.tensor(batch).shape == batch_shape

    # check the the final batch has the same dimensionf for all but the partial_batch_dim
    partial_batch = batches[-1]
    partial_batch_shape = torch.tensor(partial_batch).shape
    # remove the partial index from each and check that shape matches otherwise
    batch_shape_list = list(batch_shape)
    partial_batch_shape_list = list(partial_batch_shape)

    del batch_shape_list[-1]
    del partial_batch_shape_list[-1]

    assert batch_shape_list == partial_batch_shape_list


def test_batch_invariant_sampler_no_batch_invariant():
    samples = list(range(8))
    # What ould happen if full size batching was used
    batch_sampler2 = BatchInvariantSampler2d(samples, batch_shape=torch.Size([2, 4]))
    samples2 = torch.concat([torch.tensor(b) for b in batch_sampler2], dim=-1)

    # What will happen when batch dimension is changed
    batch_sampler1 = BatchInvariantSampler2d(samples, batch_shape=torch.Size([2, 2]))
    samples1 = torch.concat([torch.tensor(b) for b in batch_sampler1], dim=-1)

    torch.testing.assert_close(samples1, samples2)


def test_batch_invariant_sampler_not_2d():
    samples = list(range(12))

    with pytest.raises(ValueError, match=".*Only 2d batches are supported.*"):
        _ = BatchInvariantSampler2d(samples, batch_shape=torch.Size([1, 2, 3]))



import numpy as np

from axtreme.data.dataset import MinimalDataset
from axtreme.data.fixed_random_sample import FixedRandomSampler


def test_indentical_random_sample_over_multiple_iterations():
    """Check that each time the sampler is iterated over it produces identical samples.

    This is what makes the dataloader produce identical data each time it it iterated over (in its entirety).
    """
    data = np.arange(5).reshape(-1, 1)
    dataset = MinimalDataset(data)

    sampler = FixedRandomSampler(dataset, num_samples=10, seed=0, replacement=False)

    iteration1 = list(sampler)
    iteration2 = list(sampler)

    assert iteration1 == iteration2



"""Defines the interface posterior samplers are expected to adhear to."""

from collections.abc import Callable
from typing import Protocol

import torch
from botorch.posteriors import GPyTorchPosterior


class PosteriorSampler(Protocol):
    """Defines the protocol for sampler function.

    This follows the definition of the MCSampler.forward() method, but allows for simple function based implementations.

    Note:
        Function using posteriors should check if there are special `mean` and `var` methods if they support posterior
        samplers that require them (e.g UT). See note below for details.

    Notes on samplers the require special aggregation of results:

    - We use a surrogate model to estimate a Quantity of Interest (QOI), and provide uncertainty of that estimate.
      Typically it is challenging to calculate this on the posterior directly, so often we take samples of the
      posterior and calculate the value of interest using those. We can then combine those estimates to give the QOI
      mean/variance for the overall posterior.
    - Some methods (e.g UT) select posterior samples in a special way which means less samples are needed. The
      calculations for each of these samples then need to be combined in a special way to estimate the QOI
      mean/variance for the overall posterior. If such special methods are needed, they should be implement on the
      posterior sampler with the following signature.

      .. code-block:: python

        def mean(self, x: torch.Tensor) -> torch.Tensor: ...
        def var(self, x: torch.Tensor) -> torch.Tensor: ...

      Where

      - `x` is the scores (e.g QOIs) for each posterior sample.

        - It is expected to be of shape (*batch_shape*)

        .. Todo::
                Batching gets confusing here, how should we do UT of multiple batch dims? Is it only
                ever flat, should we say that here?

        - Output is a scalar.

    Todo:
        Need to fix/revisit this. Feel wrong that functions that use posterior need to check for some optional methods
        that "might" be there (and those functions are not programmatically defined as part of the protocol). Extra
        info in #132.

        .. Note::
            The functions are not on the protocol because we want MCSampler from botorch to fall into protocol
            definition.
    """

    def __call__(self, posterior: GPyTorchPosterior) -> torch.Tensor:
        """Produces samples from a posterior distibution.

        Args:
            posterior: the posterior object to be sampled

        Return:
            Tensor: shape (`*sample_shape, *b, n, m`), where:

              - `sample_shape`: the number/shape of samples produces
              - `*b, n, m`: as defined in glossary.

        """
        ...


class MeanVarPosteriorSampledEstimates(Protocol):
    """This is a protocol for classes that computes estimates using a PosteriorSampler.

    Some posterior samplers require that mean and variance of the estimates calculated by sampling from it need to be
    calculated in a special way. This protocol allows for these special methods to be easier to be used on the
    outside of the class that inherits from this protocol. One example of a case where this is useful is when using a
    posterior sampler in a QoIEstimator. For more information check issue #132.

    Example of usage:

    .. code-block:: python

        class MyClass(MeanVarPosteriorSampledEstimates):
            def __init__(self, posterior_sampler: PosteriorSampler):
                self.posterior_sampler = posterior_sampler

            def f(self, x: torch.Tensor) -> torch.Tensor:
                # Some functions that takes the samples and returns a tensor
                # The dimension representing the posterior_sampling should be preserved.
                ...

            def foo(self) -> torch.Tensor:
                #  Get a model and points x to evaluate posterior
                x: torch.Tensor = ...
                model: botorch.models.Model = ...

                # Calculating posterior
                posterior = model.posterior(x)

                # Sampling from the posterior
                samples = self.posterior_sampler(posterior)

                return self.f(samples)


        estimator = MyClass(posterior_sampler)
        result = estimator.foo()

        # Using the mean and var methods from MeanVarPosteriorSampledEstimates
        # This will use the mean and var methods from the posterior_sampler if they are available.
        mean = estimator.mean(result)
        var = estimator.var(result)
    """

    posterior_sampler: PosteriorSampler

    def mean(self, x: torch.Tensor) -> torch.Tensor:
        """Function that computes the mean of the estimates produced by using `self.posterior_sampler`.

        The dimension representing the posterior_sampling should be preserved from the sampling using
        `self.posterior_sampler` to sample a posterior until using this method.

        For many applications this method will just be using a default implementation that computes the mean.
        E.g using torch.mean(x) like in this implementation.

        However, in some special cases, it might be useful to provide a custom implementation
        to give a more accurate estimate.
        For instance if one uses UTSampler to sample the posterior
        the mean should be estimated as a weighted sum of the estimates.

        This function uses the mean method of the posterior_sampler if it is available,
        otherwise it uses the default implementation.

        Args:
            x: A tensor of the estimates with shape (num_posterior_samples,).

        Returns:
            torch.Tensor: The mean of the the estimates. Should be a scalar with shape ().
        """
        # If there is a `mean()` method in the `posterior_sampler` then use it.
        # Otherwise use the default implementation from `torch`.
        mean_func: Callable[[torch.Tensor], torch.Tensor] | None = getattr(self.posterior_sampler, "mean", None)
        return mean_func(x) if mean_func and callable(mean_func) else torch.mean(x, dim=0)

    def var(self, x: torch.Tensor) -> torch.Tensor:
        """Function that computes the variance of the estimates produced by using `self.posterior_sampler`.

        The dimension representing the posterior_sampling should be preserved from the sampling using
        `self.posterior_sampler` to sample a posterior until using this method.

        For many applications this method will just be using a default implementation that computes the variance.
        E.g using torch.var(x) like in this implementation.

        However, in some special cases, it might be useful to provide a custom implementation
        to give a more accurate estimate.
        For instance if one uses UTSampler to sample the posterior
        the variance should be estimated in a special way.

        This function uses the variance method of the posterior_sampler if it is available,
        otherwise it uses the default implementation.

        Args:
            x: A tensor of the estimates with shape (num_posterior_samples,).

        Returns:
            torch.Tensor: The variance of the the estimates. Should be a scalar with shape ().
        """
        # If there is a `var()` method in the `posterior_sampler` then use it.
        # Otherwise use the default implementation from `torch`.
        var_func: Callable[[torch.Tensor], torch.Tensor] | None = getattr(self.posterior_sampler, "var", None)
        return var_func(x) if var_func and callable(var_func) else torch.var(x, dim=0)



"""Datasets to help work with numpy files (.npy)."""

# %%
import functools
from pathlib import Path

import numpy as np
import torch
from numpy.typing import NDArray
from torch.utils.data import Dataset


@functools.lru_cache(maxsize=1)
def collect_np_file(file_path: str | Path) -> NDArray[np.float64]:  # noqa: D103
    return np.load(file_path)


class NumpyFileDataset(Dataset[torch.Tensor]):
    """Helper to work with directories of .npy data.

    Note:
        - Highly recommened to use an in memory dataset if possible. This is typically a bottleneck.
        - Using with a sequential sampler will be significantly faster because this performs rudimental cacheing.
            - Random sampling will require from disk read for EVERY datapoint. Suggest randomise the save files.

    Assumes:
        - Each row is a data point
        - Each file has the same number of datapoints within it.

    Dev:
        - Based on example here: https://pytorch.org/tutorials/beginner/data_loading_tutorial.html

    Todo:
        - This is slow compared to reading from memory. 100k dataset (3ms for memory, 20s with this dataloader)
            - This is because EVERY SINGLE datapoint requireds a new read from disk.
            - TODO: Consider different file types that might be more appropriate (row specfic access)
                - HDF5?
            - TODO: consider intergration with an imporance weight
                - Look at the example with images, can return many things
                - npz allows you to combine multiple arrays, could have soem logic if other array none, no imporance.
    Answered:
        - Is Sampler a more approapriate way of framing this?
            - No. The samples class just take an existing dataset and shuffles it. Like shuffle in dataloader.
    """

    def __init__(self, root_dir: str | Path) -> None:
        """Initialise the Dataset.

        Note:
            Data should be loaded lazily (in `__getitem__`, not here)

        Arguments:
            root_dir (string): Directory with .npy files
        """
        self.root_dir = root_dir
        self.paths = list(Path(root_dir).glob("*.npy"))
        self.row_per_file = np.load(self.paths[0]).shape[0]

    def __len__(self) -> int:  # noqa: D105
        # This doesn't do anything to control when samplers will stop sampling. Just tells the range of valid indexs
        # RandomSampler might sample many more points than is in this dataset (even if replace = False)
        # Sequential will just iterate though once

        # Basically: When stop drawing data is based on the logic in the sampler.
        return self.row_per_file * len(self.paths)

    def __getitem__(self, idx: int) -> torch.Tensor:
        """In this function you want to get only the index of data required.

        Args:
            idx: what index in the dataset should be returned

        Returns:
            A tensor representing a single datapoint.

        Note:
            See the example linked in class docs, can return whatever type of object would like.
            This could be useful for importance sampling.
        """
        file_idx = idx // self.row_per_file
        row_idx = idx % self.row_per_file

        file = collect_np_file(self.paths[file_idx])
        sample = file[row_idx, :]
        # NOTE: this should NOT be on GPU. That is done lazily as needed in the training loop.
        # If put on GPU now Dataloader works can't build an effecient GPU queue.
        return torch.from_numpy(sample)



# pyright: reportUnnecessaryTypeIgnoreComment=false

# sourcery skip: no-conditionals-in-tests

from collections.abc import Callable

import matplotlib.pyplot as plt
import pytest
import torch
from botorch.models import SingleTaskGP
from botorch.models.model import Model
from botorch.optim import optimize_acqf

from axtreme.acquisition.qoi_look_ahead import (
    QoILookAhead,
    average_observational_noise,
    closest_observational_noise,
    conditional_update,
)
from axtreme.plotting.gp_fit import plot_1d_model
from axtreme.qoi.qoi_estimator import QoIEstimator
from axtreme.utils.gradient import is_smooth_1d
from tests.acquisition.helpers import check_posterior, check_same_model_data
from tests.helpers import (
    single_task_fixed_noise_m_1,
    single_task_fixed_noise_m_1_outcome_transform,
    single_task_homo_noise_m_1,
)


def test_get_closest_observational_noise():
    new_points = torch.tensor([[0.1, 0.1], [0.16, 0.16], [0.4, 0.4], [0.4, 0.4]])
    train_x = torch.tensor([[0.1, 0.1], [0.2, 0.2], [0.3, 0.3]])
    train_yvar = torch.tensor([[0.11], [0.22], [0.33]])

    # Run test
    actual_output = closest_observational_noise(new_points, train_x, train_yvar)

    torch.testing.assert_close(actual_output, torch.tensor([[0.1100], [0.2200], [0.3300], [0.3300]]))


def test_get_average_observation_noise():
    """check the right shaped output is created"""
    new_points = torch.tensor([[0.1, 0.1], [0.16, 0.16], [0.4, 0.4], [0.4, 0.4]])
    # m = 2
    train_yvar = torch.tensor([[0.11, 0.12], [0.22, 0.23], [0.36, 0.37]])

    actual_output = average_observational_noise(new_points, None, train_yvar)

    expected_result = torch.tensor([[0.23, 0.24], [0.23, 0.24], [0.23, 0.24], [0.23, 0.24]])
    torch.testing.assert_close(actual_output, expected_result)


@pytest.mark.parametrize(
    "method",
    [
        (average_observational_noise),
        pytest.param(closest_observational_noise, marks=pytest.mark.xfail(reason="First derivate is not smooth")),
    ],
)
def test_obersvational_noise_getting_is_smooth(
    method: Callable[[torch.Tensor, torch.Tensor, torch.Tensor], torch.Tensor],
) -> None:
    """Check the function getting yvar_new is smooth, which is important for optimisation.

    _get_fantasy_observation_noise using the an underlying helper function to get the yvar_new.
    The helper need to be a smooth function if all optimisation tools are expected to be useable.
    """
    new_points = torch.linspace(0, 1, 100).reshape(-1, 1)

    train_x = torch.tensor([[0.1], [0.3], [0.6], [0.9]])
    train_yvar = torch.tensor([[0.1], [0.9], [0.1], [0.9]])

    yvar_new_function = method(new_points, train_x, train_yvar)

    _ = is_smooth_1d(new_points.flatten(), yvar_new_function.flatten())


@pytest.mark.parametrize(
    "b,n,d,m, expected_output",
    [  # Test without batching
        (None, 5, 2, 1, torch.tensor([0, 1, 2, 3, 4])),
        # target dim 2
        (None, 5, 2, 2, torch.tensor([0, 2, 4, 6, 8])),
        # add batch dim
        (2, 5, 2, 2, torch.tensor([[0, 2, 4, 6, 8], [10, 12, 14, 16, 18]])),
    ],
)
def test_batch_lookahead_correct_reshape(b: None | int, n: int, d: int, m: int, expected_output: torch.Tensor):
    """Check that the reshape process used to iterate through lookahead keeps values in the right place.

    Args:
        - b: the batch dimenstion.
        - n: the number of points
        - d: dimension of x points
        - m: number of tergets
        - expected_output: expected output of the function.
    """
    ### set up the objects
    # Set up the object
    # TODO @ClaasRostock: Not sure if should be using mock/patch libs here. seemed like overkill.
    acqf = QoILookAhead(model=None, qoi_estimator=None)  # type: ignore[arg-type]

    # mock the lookahead function. Simply return the first dimension in y
    def lookahead(x_point: torch.Tensor, y_point: torch.Tensor, yvar_point: torch.Tensor | None) -> torch.Tensor:  # noqa: ARG001
        return y_point[0]

    acqf.lookahead = lookahead  # type: ignore[method-assign]

    # Create the inputs
    if b is None:
        # (*b,n,d)
        x_points = torch.ones(n, d)
        # (*b,n,m)
        y_points = torch.arange(n * m).reshape(n, m)
        yvar_points = torch.ones_like(y_points) * 0.1
    else:
        x_points = torch.ones(b, n, d)
        # (*b,n,m)
        y_points = torch.arange(b * n * m).reshape(b, n, m)
        yvar_points = torch.ones_like(y_points) * 0.1

    actual_result = acqf._batch_lookahead(x_points, y_points, yvar_points)

    torch.testing.assert_close(actual_result, expected_output)


# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
def test_conditional_update():
    """Test connection to underling `condition_on_observation`.

    NOTE: `condition_on_observation` is tested more extensively in `test_qoi_look_ahead_external.py`
    """
    train_x = torch.tensor([[0.1], [0.2], [0.3]])
    train_y = torch.sin(train_x)
    train_yvar = torch.ones_like(train_y) * 0.1

    model_original = SingleTaskGP(train_x, train_y, train_yvar)

    # Generate new training data to use
    new_x = torch.tensor([[0.5]])  # (n,d)
    new_y = torch.tensor([[0.1]])
    new_yvar = torch.ones_like(new_y) * 0.1

    # New conditioned model
    model_conditioned = conditional_update(model_original, X=new_x, Y=new_y, observation_noise=new_yvar)

    # Expected model
    x_total = torch.vstack([train_x, new_x])  # (n+1,d=1)
    y_total = torch.vstack([train_y, new_y])  # (n+1,n_targets)
    yvar_total = torch.vstack([train_yvar, new_yvar])  # (n+1,n_targets)
    model_expected = SingleTaskGP(x_total, y_total, yvar_total)

    # Check the internal data is the same
    check_same_model_data(model_actual=model_conditioned, model_expected=model_expected)
    # check that they both make the same preditiontion
    x = torch.tensor([[0.4], [0.7]])
    expected_posterior = model_expected.posterior(x)
    conditioned_posterior = model_conditioned.posterior(x)
    check_posterior(conditioned_posterior, expected_posterior)


@pytest.mark.parametrize(
    "model, x, y, observation_noise, error_message",
    [
        pytest.param(
            single_task_homo_noise_m_1(),
            None,
            None,
            torch.tensor([0.1]),
            "observation_noise is not supported",
            id="Homoskedatic model recieved observation noise",
        ),
        pytest.param(
            single_task_fixed_noise_m_1(),
            None,
            None,
            None,
            "requires observation_noise",
            id="Fixed noise model missing observation noise",
        ),
        pytest.param(
            single_task_homo_noise_m_1(),
            torch.ones(1, 1, 1),
            torch.ones(1, 1, 1),
            None,
            "batch shape",
            id="Batched data doesn't match non-batched model",
        ),
    ],
)
def test_conditional_update_exception_checking(
    model: SingleTaskGP,
    x: None | torch.Tensor,
    y: None | torch.Tensor,
    observation_noise: None | torch.Tensor,
    error_message: str,
):
    """Check the differnet warning conditions are approapriately checked.

    Args:
        model: the model being updated.
        x: The additional points to add. None here if not involved with triggering the error
        y: The additional points to add. None here if not involved with triggering the error
        observation_noise: yvar of the additional points to add. None here if not involved with triggering the error
        error_message: substring expected in the error message.
    """
    with pytest.raises(Exception, match=error_message):
        _ = conditional_update(model, X=x, Y=y, observation_noise=observation_noise)  # type: ignore[arg-type]


def test_conditional_update_warning_checking_input_output_transforms():
    model = single_task_fixed_noise_m_1_outcome_transform()
    # Note: this is not important, just to prevent other error from failing so we can get to the one we want
    data = torch.tensor([[0.1]])

    with pytest.warns(UserWarning, match="outcome transforms"):
        _ = conditional_update(model, X=data, Y=data, observation_noise=data)


def test_forward_grad():
    r"""Test the acquisition function correctly handles gradient information.

    This is currently not tested so function should raise not yet implemented error.

    Todo:
    - Make simple test like `test_optimise_dumb_qoi` but with gradient turned on.
    - make more sophisticated tests:
        - These might cover:
            - what happens in the optimisation when there is noise in the QoI.
    - Existing material that can serve as a base.
        -`explorations\qoi_look_ahead_gradient.ipynb` to build gradient tests. This covers:
            - ploting of the acqusition function gradient at a grid of points.
            - Running with Ax with and withotu gradient to see the difference.
        - `explorations\dev_ks\qoi_acquisition\qoi_acquisition_first_trial.ipynb`
            - rough example optimising a problem with gradient through Ax.

    """
    acqf = QoILookAhead(model=None, qoi_estimator=None)  # type: ignore[arg-type]
    x = torch.ones(1, 1, 1, requires_grad=True)

    with pytest.raises(NotImplementedError):
        acqf(x)


# Would be nice to test GPU, but there is not flag so this is difficult.
# The underling module has no params to check either.


@pytest.mark.integration
# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
def test_acquisition_function_is_smooth(*, visual_inspect: bool = False):
    """Assuming the requirement detailed in the the class docs are met, the acquisition fucntion produced should be
    smooth.
    """

    train_x = torch.tensor([[0.1], [0.5], [0.9]])
    train_y = torch.zeros_like(train_x)
    train_yvar = torch.tensor([[0.9], [0.1], [0.9]])
    model = SingleTaskGP(train_x, train_y, train_yvar)

    if visual_inspect:
        _ = plot_1d_model(model)
        plt.pause(5)  # required to be able to see the plot when running in pytest

    class DummyQoI(QoIEstimator):
        """Simply QoI that only cares about reducing the variance at one point.

        Only cares about the variance at `x=.5` The var method is then just a pass through of the value returned by x.

        The output doesn't truely conform with the QoI interface, but is suffecient for this test.
        """

        def __call__(self, model: Model) -> torch.Tensor:
            x = torch.tensor([[0.5]])
            with torch.no_grad():
                posterior = model.posterior(x)

            return posterior.variance.flatten()  # type: ignore  # noqa: PGH003

        def var(self, x: torch.Tensor) -> torch.Tensor:
            return x[0]

    dummy_qoi = DummyQoI()
    acqf = QoILookAhead(model, qoi_estimator=dummy_qoi)

    x = torch.linspace(0, 1, 200).reshape(-1, 1, 1)
    scores = acqf(x)

    if visual_inspect:
        _ = plt.plot(x.flatten(), scores.flatten())
        plt.pause(5)

    _ = is_smooth_1d(x.flatten(), scores.flatten())


# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
@pytest.mark.integration
# TODO(sw 2024-11-27): This should be updated to use the default params registered in qoi_look_ahead.py
def test_optimise_dumb_qoi(*, visual_inspect: bool = False):
    """Make a QoI that only cares about reducing the variance as a single point.

    Check that the optimisation can find that point.

    Args:
        visual_inspect: Flag to create visualisation of test for manual inspection.

    Note:
        We could also test more complicated surfaces and show they work. Testing if optimiser setting work on a
        realistic problem surface is consider system testing. Testing if optimiation setting work on arbitrary
        problem surface should be in unittests of the optimiser (with arbirary functions) rather than here.
    """
    train_x = torch.tensor([[0.1], [0.9]])
    train_y = torch.zeros_like(train_x)
    train_yvar = torch.ones_like(train_y) * 0.1
    model = SingleTaskGP(train_x, train_y, train_yvar)

    if visual_inspect:
        _ = plot_1d_model(model)
        plt.pause(5)  # required to be able to see the plot when running in pytest

    class DummyQoI(QoIEstimator):
        """Simply QoI that only cares about reducing the variance at one point.

        Only cares about the variance at `x=.5` The var method is then just a pass through of the value returned by x.

        The output doesn't truely conform with the QoI interface, but is suffecient for this test.
        """

        def __call__(self, model: Model) -> torch.Tensor:
            x = torch.tensor([[0.5]])
            with torch.no_grad():
                posterior = model.posterior(x)

            return posterior.variance.flatten()  # type: ignore  # noqa: PGH003

        def var(self, x: torch.Tensor) -> torch.Tensor:
            return x[0]

    dummy_qoi = DummyQoI()
    acqf = QoILookAhead(model, qoi_estimator=dummy_qoi)

    candidate, result = optimize_acqf(
        acqf,
        bounds=torch.tensor([[0.0], [1.0]]),
        q=1,
        # This is how many different start location will be tried by the optimiser
        num_restarts=5,
        raw_samples=100,
        # Key parameter to control if optimisation should use gradient from the acquisiton fucntion
        options={"with_grad": False},  # True by default
    )

    # Could run the optimisation longer if we require it to be more precise. This is considered approapriate for a test.
    torch.testing.assert_close(candidate, torch.tensor([[0.5]]), rtol=0, atol=1e-3)



"""Allows retuning batches of data of an arbitrary dimension."""
# pyright: reportUnnecessaryTypeIgnoreComment=false

from collections.abc import Iterable, Iterator

import torch
from torch.utils.data import Sampler


class MultiBatchSampler(Sampler[list[int]]):
    """Reads the entire sampler into batches of shape `batch_shape`.

    The final batch may not have enough samples to completely fill the batch shape. Behaviour is then as follows:

    - If `partial_batch_dim` is not None, attempt to batch samples allowing this dim to be variable in size. E.g:
        >>> batch_shape = torch.Size([3, 5])
        >>> partial_batch_index = -1
        try: remaining_samples.view([3,-1])

    This check is performed up front and the sampler will throw and error.

    Warning:
        - If the batch shape changes, even in the `partial_batch_index` dimension, data will be returned in a different
          order.
        - See BatchSampler2d for details.

    Todo:
        - Determing the right approach for handling partial batch that requires more than one parital index.
          e.g fit 5 items into batch_shape = torch.Size([3,2])
        - Perhaps return a new batch that is no bigger than the original in any dimension. This batcher is likely
          for gp throughput, so then the batches only have performance not logical meaning.
    """

    def __init__(
        self,
        sampler: Sampler[int] | Iterable[int],
        batch_shape: torch.Size,
        partial_batch_dim: None | int = -1,
    ) -> None:
        """Allows you to produces batchs of arbitrary shape.

        Args:
            sampler (Sampler[int] | Iterable[int]): Sampler (e.g RandomSampler or SerquentialSampler).

                - See torch DataLoader implementation for an examples

            batch_shape (torch.Size): The batch shape created of the underling sample.
            partial_batch_dim: Dimension of the batch that can be partially filled if there is not enough samples in
                sampler.

                - Currently only one dimension can be partially filled
                - if None, no dimension is allowed to be partially filled
        """
        self.sampler = sampler
        self.batch_shape = batch_shape
        self.partial_batch_dim = partial_batch_dim

        # Upfront check the the batch size is correct - don't want to fail after spending lots of time training
        # NOTE: see note about calling len(sampler) in method __len__
        data_len: int = len(sampler)  # type: ignore[arg-type]
        partial_batch_data_len = data_len % batch_shape.numel()
        if partial_batch_data_len != 0:
            if partial_batch_dim is None:
                msg = (
                    f"Final batch will have {partial_batch_data_len} items,"
                    f" which does not fit into batch shape {batch_shape}"
                )
                raise ValueError(msg)
            # Check the data perfectly fits into the specified partial batch shape.
            shape = list(batch_shape)
            shape[partial_batch_dim] = -1
            final_batch_shape = torch.Size(shape)
            if partial_batch_data_len % final_batch_shape.numel() != 0:
                msg = (
                    f"Final batch will have {partial_batch_data_len} items,"
                    f" which does not fit into batch shape {final_batch_shape}"
                )
                raise ValueError(msg)

    def __len__(self) -> int:  # noqa: D105
        # NOTE: len is not a guaranteed method on Sampler. but all the subclasses we use have it.
        #       see note about calling len(sampler) in method __len__
        data_len: int = len(self.sampler)  # type: ignore[arg-type]
        return (data_len + self.batch_shape.numel() - 1) // self.batch_shape.numel()

    def __iter__(self) -> Iterator[list[int]]:
        """Returns a nested lists of the shape specified by `batch_shape`."""
        # Do all batches except the last, which might be partial batch
        sample_iter = iter(self.sampler)
        for _ in range(len(self) - 1):
            # the follow with throw StopIteration if there is not enough data. This should flow to the user.
            batch = [next(sample_iter) for _ in range(self.batch_shape.numel())]
            yield torch.tensor(batch).view(self.batch_shape).tolist()

        # determine the shape of the final batch
        final_batch_shape = self.batch_shape
        if self.partial_batch_dim is not None:
            shape = list(self.batch_shape)
            shape[self.partial_batch_dim] = -1
            final_batch_shape = torch.Size(shape)

        yield torch.tensor(list(sample_iter)).reshape(final_batch_shape).tolist()



"""Utility functions for working with botorch models."""

import torch
from botorch.models import SingleTaskGP
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood, GaussianLikelihood


def get_training_data_singletaskgp(model: SingleTaskGP) -> torch.Tensor:
    """Extracts the training data from SIngleTaskGP in consistent form.

    As per Botorch documentation, "SingleTaskGP Modules same training data for all outputs" (`BoTorch docs <https://botorch.org/docs/models#single-task-gps>`_)
    As such we are not concerned with batch dimenstion. Gpytorch stores training data in this format:

        - if m = 1: (`*b`,n,d)
        - if m > 1: (`*b`,m,n,d)

    Args:
        model: the model to collect the training data from.

    Returns:
        training_inputs: (n,d) A torch.Tensor of the training_inputs

    NOTE: The interface defined here is only available for SingleTaskGPs, as other model can contain different input
        data for each batch.
    """
    input_data_tuple = model.train_inputs
    assert input_data_tuple is not None, f"model.train_inputs is None: {input_data_tuple}"
    assert len(input_data_tuple) > 0, f"model.train_inputs is empty: {input_data_tuple}"
    if len(input_data_tuple) > 1:
        msg = f"model.train_inputs has more than one element: {input_data_tuple}"
        raise ValueError(msg)

    # Get to the data
    batched_input_data = input_data_tuple[0]

    batch_shape = batched_input_data.shape[:-2]
    index = [0] * len(batch_shape) + [slice(None), slice(None)]
    input_data = batched_input_data[index]
    return input_data


def get_target_noise_singletaskgp(model: SingleTaskGP) -> torch.Tensor:
    """A helper to extract the training varaince from a botorch model, and return it in a consistent format.

    Args:
        model: A SingleTaskGP model.

    Returns:
        noise (variance): A torch.Tensor of the noise with shape `(*b, n,m)` where b is an optional batch dimention, `n`
        is the number of observations, `m` is the number of targets. If homoskedatic noise is used shape will be
        `(*b, n=1,m)`.

    NOTE: Approach is based on FantasizeMixin.fantasize.
    NOTE: Currently unclear if this interface extends to othe type of model.
    """
    # Fixed and Homoskedatic noise
    if isinstance(model.likelihood, FixedNoiseGaussianLikelihood | GaussianLikelihood):
        # this chunk in based on FantasizeMixin.fantasize
        if model.num_outputs > 1:
            # change from (*b, m, n) to (*b,n,m)
            # fixed noise: (*b, m, n) to (*b,n,m)
            # homoskedatic noise: (*b, m, n = 1) to (*b,n = 1,m)
            observation_noise = model.likelihood.noise.transpose(-1, -2)
        else:
            # Change from (*b,n) to (*b,n,m=1)
            # fixed noise: (*b,n = 1) to (*b,n = 1,m=1)
            # homoskedatic noise: (*b,n = 1) to (*b,n = 1,m=1)
            observation_noise = model.likelihood.noise.unsqueeze(-1)

        return observation_noise

    msg = (
        "Expected model.likilihood of `FixedNoiseGaussianLikelihood` or `GaussianLikelihood`,"
        f" got {type(model.likelihood.noise)}"
    )
    raise NotImplementedError(msg)



import pytest
import torch
from botorch.models.gp_regression import SingleTaskGP
from tests.helpers import (
    single_task_fixed_noise_m_1,
    single_task_fixed_noise_m_2,
    single_task_homo_noise_m_1,
    single_task_homo_noise_m_1_b3,
    single_task_homo_noise_m_2,
    single_task_homo_noise_m_2_b3,
)

from axtreme.utils.model_helpers import get_target_noise_singletaskgp, get_training_data_singletaskgp


@pytest.mark.parametrize(
    "model, expected_shape",
    [
        (single_task_homo_noise_m_1(), torch.Size([10, 1])),
        (single_task_homo_noise_m_1_b3(), torch.Size([10, 1])),
        (single_task_homo_noise_m_2(), torch.Size([10, 1])),
        (single_task_homo_noise_m_2_b3(), torch.Size([10, 1])),
    ],
)
def test_get_training_data_singletaskgp(model: SingleTaskGP, expected_shape: torch.Size):
    """Test training data of consistent format can be extracted from a model.

    Varient that effect the underling shape (b is batch shape, m is number of targets):
     - b = 0, m = 1
     - b = 3, m = 1
     - b = 0, m = 2
     - b = 3, m = 2

    Functionality is invariate to noise type, so that is not tested.

    NOTE: n0n=10 is dependant on the training data used in the model. If this changes the test will break.
    """
    result_shape = get_training_data_singletaskgp(model).shape

    torch.testing.assert_close(result_shape, expected_shape)


@pytest.mark.parametrize(
    "model, expected_shape",
    [
        # Homoskedastic
        (single_task_homo_noise_m_1(), torch.Size([1, 1])),
        (single_task_homo_noise_m_1_b3(), torch.Size([3, 1, 1])),
        (single_task_homo_noise_m_2(), torch.Size([1, 2])),
        (single_task_homo_noise_m_2_b3(), torch.Size([3, 1, 2])),
        # Fixed noise
        (single_task_fixed_noise_m_1(), torch.Size([10, 1])),
        (single_task_fixed_noise_m_2(), torch.Size([10, 2])),
    ],
)
def test_get_target_noise(model: SingleTaskGP, expected_shape: torch.Size):
    """Tests that different noise and shape types are treated constitenly are treated consistently."""
    result_shape = get_target_noise_singletaskgp(model).shape

    torch.testing.assert_close(result_shape, expected_shape)



import numpy as np

from axtreme.data import MinimalDataset
from axtreme.data.sizable_sequential_sample import SizableSequentialSampler


def test_SizableSequentialSampler_less_sample_than_data():
    minimal_dataset = MinimalDataset(np.arange(8))

    seq_samper = SizableSequentialSampler(minimal_dataset, num_samples=4)
    assert list(seq_samper) == [0, 1, 2, 3]


def test_SizableSequentialSampler_from_data():
    minimal_dataset = MinimalDataset(np.arange(8))

    seq_samper = SizableSequentialSampler(minimal_dataset)
    assert list(seq_samper) == [0, 1, 2, 3, 4, 5, 6, 7]


def test_SizableSequentialSampler_more_sample_than_data():
    minimal_dataset = MinimalDataset(np.arange(8))

    seq_samper = SizableSequentialSampler(minimal_dataset, num_samples=10)
    assert list(seq_samper) == [0, 1, 2, 3, 4, 5, 6, 7, 0, 1]



"""3D histogram plotting function.

Plotly does not have a plotly.graph_objects.Histogram3d object, this is an attempt to remedy that.
The closest is plotly.graph_objects.Histogram2d, but that creates a 2D heatmap.

See Also:
- https://github.com/serge-tochilov/barchart3d-plotly
- https://github.com/AymericFerreira/Plotly_barchart3D
- https://community.plotly.com/t/how-to-do-a-3d-bar-chart-if-possible/32287/5
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any

import numpy as np
import plotly.graph_objects as go


def histogram3d(  # noqa: PLR0913
    samples: np.ndarray[tuple[int, int], np.dtype[np.float64]],
    *,
    n_bins: int = 20,
    bounds: Sequence[tuple[float, float]] | None = None,
    density: bool = True,
    flatshading: bool = True,
    show: bool = False,
    mesh3d_kwargs: dict[str, Any] | None = None,
    layout_kwargs: dict[str, Any] | None = None,
) -> go.Figure:
    """Generate a 3D figure of a 2D histogram of the input samples.

    Computes a 2D histogram that is then plotted in 3D as a 3D bar chart (confusing naming...).

    Args:
        samples: An array of shape (n_samples, 2) of samples to compute the histogram of.
        n_bins: The number of bins to use in each dimension.
        bounds: An array of shape (2, 2) of the bounds of the histogram: ((x_min, x_max), (y_min, y_max)). If None, the
            bounds are handled by numpy.histogramdd.
        density: If True, the histogram is normalized to form a probability density function.
        flatshading: If True, the 3D bars are shaded smoothly.
        show: If True, the figure is displayed.
        mesh3d_kwargs: A dictionary of keyword arguments to pass to the plotly Mesh3d object.
        layout_kwargs: A dictionary of keyword arguments to pass to the update_layout method of the figure.

    Returns:
        A plotly figure of the 3D histogram. If show is True, the figure is also displayed.
    """
    # set default values
    if mesh3d_kwargs is None:
        mesh3d_kwargs = {}
    if layout_kwargs is None:
        layout_kwargs = {}

    if bounds is None:
        bounds = [
            (float(samples[:, 0].min()), float(samples[:, 0].max())),
            (float(samples[:, 1].min()), float(samples[:, 1].max())),
        ]

    # compute the histogram
    hist, edges = np.histogramdd(
        samples,
        bins=n_bins,
        range=bounds,
        density=density,
    )
    x_edges = edges[0]
    y_edges = edges[1]

    # construct the 3D bars manually
    fig = go.Figure()

    # add bars for each bin
    for i in range(n_bins):
        for j in range(n_bins):
            z_value = hist[i, j]
            if z_value > 0:  # only plot non-empty bins
                x_min = x_edges[i]
                x_max = x_edges[i + 1]
                y_min = y_edges[j]
                y_max = y_edges[j + 1]

                _ = fig.add_trace(
                    go.Mesh3d(
                        x=[x_min, x_min, x_max, x_max, x_min, x_min, x_max, x_max],
                        y=[y_min, y_max, y_max, y_min, y_min, y_max, y_max, y_min],
                        z=[0, 0, 0, 0, z_value, z_value, z_value, z_value],
                        alphahull=0,
                        **{
                            "intensity": [0, 0, 0, 0, z_value, z_value, z_value, z_value],
                            "flatshading": flatshading,
                            "coloaxtremeis": "coloaxtremeis",
                            **mesh3d_kwargs,
                        },
                    ),
                )

    _ = fig.update_layout(
        **{
            "scene": {
                "xaxis": {"title": "X1", "range": bounds[0]},
                "yaxis": {"title": "X2", "range": bounds[1]},
                "zaxis": {"title": "Density" if density else "Count"},
            },
            **layout_kwargs,
        },
    )

    if show:
        fig.show()

    return fig


def histogram_surface3d(
    samples: np.ndarray[tuple[int, int], np.dtype[np.float64]],
    *,
    n_bins: int = 20,
    bounds: Sequence[tuple[float, float]] | None = None,
    density: bool = True,
    show: bool = False,
    surface3d_kwargs: dict[str, Any] | None = None,
    layout_kwargs: dict[str, Any] | None = None,
) -> go.Figure:
    """Generate a 3D figure of a 2D histogram of the input samples.

    Computes a 2D histogram that is then plotted in 3D as a surface plot.

    Args:
        samples: An array of shape (n_samples, 2) of samples to compute the histogram of.
        n_bins: The number of bins to use in each dimension.
        bounds: An array of shape (2, 2) of the bounds of the histogram: ((x_min, x_max), (y_min, y_max)). If None, the
            bounds are handled by numpy.histogramdd.
        density: If True, the histogram is normalized to form a probability density function.
        show: If True, the figure is displayed.
        surface3d_kwargs: A dictionary of keyword arguments to pass to the plotly Surface3d object.
        layout_kwargs: A dictionary of keyword arguments to pass to the update_layout method of the figure.

    Returns:
        A plotly figure of the 3D histogram. If show is True, the figure is also displayed.
    """
    # set default values
    if surface3d_kwargs is None:
        surface3d_kwargs = {}
    if layout_kwargs is None:
        layout_kwargs = {}

    if bounds is None:
        bounds = [
            (float(samples[:, 0].min()), float(samples[:, 0].max())),
            (float(samples[:, 1].min()), float(samples[:, 1].max())),
        ]

    # compute the histogram
    hist, edges = np.histogramdd(
        samples,
        bins=n_bins,
        range=bounds,
        density=density,
    )
    x_edges = edges[0]
    y_edges = edges[1]

    fig = go.Figure(
        data=[
            go.Surface(
                x=x_edges,
                y=y_edges,
                z=hist.T,
                **surface3d_kwargs,
            ),
        ],
    )

    _ = fig.update_layout(
        **{
            "scene": {
                "xaxis": {"title": "X1", "range": bounds[0]},
                "yaxis": {"title": "X2", "range": bounds[1]},
                "zaxis": {"title": "Density" if density else "Count"},
            },
            **layout_kwargs,
        },
    )

    if show:
        fig.show()

    return fig



"""Contains objects and helpers for creating data unit testing."""

from collections.abc import Sequence
from typing import Any, TypeVar, cast

import torch
from numpy.typing import NDArray
from torch.utils.data import Dataset

# TODO(sw): Dataset makes the type covariant (e.g TypeVar('T_co', covariant=True)).
# Struggling to understand contravariate/covariant/invariant. Come back to this.
T = TypeVar("T", Sequence[Any], NDArray[Any], torch.Tensor)


class MinimalDataset(Dataset[T]):
    """Creates a Dataset over a list-like data source.

    Note:
        `list`, `np.array` or `tensor` forfill the `__getitem__` and `__len__` requirement directly - but this makes
        them conform with the __add__ behaviour defined for datasets.
    """

    def __init__(self, data: Sequence[T] | T) -> None:  # . D417
        """Creates a Dataset over a list-like data source.

        Args:
            data: Supports being indexed (supports __getitem__) and len, where index values are considered datapoints.
            - e.g (n,d) numpy dataset, a List[datapoint].
            - Datapoints should be compatible with DataLoader.
                - General work with numerics and matrixes
                - Specifics of what DataLoader consume from datasets see `torch._utils.collate.default_collate`

        Examples:
            >>> data = [1, 2, 3]
            >>> ds = MinimalDataset(data)
            >>> ds[0]
            1

            >>> data = np.array([[1, 2, 3], [4, 5, 6]])
            >>> ds = MinimalDataset(data)
            >>> ds[0]
            1p.array([1,2,3])

            >>> data = torch.tensor([[1], [2]])
            >>> ds = MinimalDataset(data)
            >>> ds[0]
            torch.tensor([1])
        """
        self.data: Sequence[T] | T = data

    def __len__(self) -> int:  # noqa: D105
        return len(self.data)

    def __getitem__(self, idx: int) -> T:
        """Returns the item at a given index.

        Return: Elements of the list like object.

        Note:
            Datapoints should be compatible with DataLoader. General we work with numerics and matrixes
            (e.g tuples, dicts, int, float, numpy, torch). DataLoader internally convers to torch object.
            For specific DataLoader transformation/consumption see `torch._utils.collate.default_collate`.
        """
        return cast(T, self.data[idx])



"""Base class for sampling methods that ignore covariance between different input points.

Todo: TODO
- For MultiTask models (e.g GPs with multiple targets that are not indpendant), this covariance should be considered.
- Currently ALL covariance is ignored.
"""

from abc import ABC, abstractmethod

import linear_operator
import torch
from botorch.posteriors import GPyTorchPosterior, Posterior
from botorch.sampling import MCSampler
from gpytorch.distributions.multitask_multivariate_normal import MultitaskMultivariateNormal
from gpytorch.distributions.multivariate_normal import MultivariateNormal


class IndependentMCSampler(MCSampler, ABC):
    r"""Abstract base class for MCSamplers that independantly apply the same set of base samples at each x point.

    This should be used when you want to ignore the covariance between different x points, and sample the X output
    space independently.

    Note:
        Dimensions of the posterior are described using botorch notation (detailed in glassary.md) where:

        - \*b: batch shape (can have arbitrary dimensionality).
        - n: the number of x input points.
        - m: dimensionality of targets space.

        For example:

        - MultivariateNormal: (\*b, n)
        - MultitaskMultivariateNormal: (\*b, n, m)

    *Dev Notes*:
      - Is the shape of the base samples up to us to define, or are there some other parts of the system that expect
        MCSampler to have certain shape.

          Answer: MCSampler does not define base_sample - we can do what we want.

      - MultivariateNormal and MultitaskMultivariateNormal both requires shape (\*sample_shape, \*posterior.shape()
        for `base_samples` in `posterior.rsample_from_base_samples`

        - For MultitaskMultivariateNormal: posterior.shape() = (\*b,n,m)
        - For MultivariateNormal: posterior.shape() = (\*b,n)

    **Design decision**: we decide to make m explicit throughout.

      - e.g base samples will always have shape (\*sample_shape,m)
      - This makes our code cleaner. When then convert back to implicit m dimension when calling
        `posterior.rsample_from_base_samples`
    """

    def forward(self, posterior: Posterior) -> torch.Tensor:
        r"""Draw samples from the posterior, treating each of the n input points as independant.

        Args:
          posterior: The posterior which should be sampled.

        Returns:
          Returns a samples of the posterior with shape: (\*sample_shape, \*b, n, m) where:
          - \*b: is the posterior batch shape
          - n: number of input points in posterior
          - m: dimensionality of target

          NOTE: While MultitaskMultivariateNormal and MultivariateNormal have different shapes ((\*b,n,m) and (\*b,n))
          the output of this function is always of shape (\*sample_shape, \*b, n, m).
          This is consistent with the behaviour of `MCSampler`.
        """
        if not isinstance(posterior, GPyTorchPosterior):
            msg = f"Currently only supports posterior of type GPyTorchPosterior. Got {type(posterior)}"
            raise NotImplementedError(msg)

        # This removes covariance from the posterior because when `posterior.rsample_from_base_samples` is applied the
        # base samples are passed through the tranformation: \mu + L*base_samples. Where
        #   - \mu: is the mean
        #   - L: is the "root" of the convariance matrix.
        # If there is covariance in L, independant base_samples will produce covariate results
        edited_posterior = diagonalize_distribution(posterior)

        # Produces base_samples of size (*sample_shape, m)
        self._construct_base_samples(edited_posterior)

        extended_base_samples = self._extend_base_samples(self.base_samples, posterior)

        # required base sample shape:
        #  - (*b,n,m) for MultitaskMultivariateNormal
        #  - (*b,n)  for MultivariateNormal
        return edited_posterior.rsample_from_base_samples(
            sample_shape=self.sample_shape, base_samples=extended_base_samples
        )

    @abstractmethod
    def _construct_base_samples(self, posterior: GPyTorchPosterior) -> None:
        """Build the base samples that should be used for sampling the GP output produced by a single input (x).

        `IndependentMCSampler`s apply the same base samples to every x in the posterior (independantly). The base
        samples to be used are define in this method.

        Return:
            tensor of shape (*sample_shape, m).
            NOTE m dimension should be present even if posterior.distribution is MultivariateNormal.

        This is then store in self.register_buffer("base_samples", base_samples)
        """
        raise NotImplementedError("Subclasses must implement this method")

    @classmethod
    def _extend_base_samples(cls, base_samples: torch.Tensor, posterior: GPyTorchPosterior) -> torch.Tensor:
        """Extend base samples to the shape required for `posterior.rsample_from_base_samples`.

        Args:
            base_samples: of shape (*sample_shape, m)
            posterior: The distibution where the base samples are intended to be applied.

        Return:
            A view of base_smaples of shape (*sample_shape, *posterior.shape()).
                - For MultitaskMultivariateNormal: (*sample_shape, *b, n, m)
                - For MultivariateNormal: (*sample_shape, *b, n)
        """
        # In order to handle the different shaped posterior more easily, we convert them both to (*b,n,m).
        # MultivariateNormal has shape (*b,n) (with an implicit m=1). We convert result back to implicit m=1 at the end.
        explicit_posterior_shape = cls._explicit_posterior_shape(posterior)

        # We eventually want to broadcast this to each n (*sample_shape, *b, n, m)
        # By turning base samples into the following shape it becomes possible: (*sample_shape, *1, 1, m)
        batchable_shape = (
            base_samples.shape[:-1]  # the sample_shape
            + torch.Size([1] * len(explicit_posterior_shape[:-1]))  # *b and n dimension of posterior
            + explicit_posterior_shape[-1:]  # m dimension
        )

        # As all the dimensions being added are 1, the underlying sample points don't change.
        batchable_base_samples = base_samples.view(batchable_shape)

        # build the explicity shape of (*sample_shape, *b, n, m)
        expanded_explicit_base_sample = batchable_base_samples.expand(
            base_samples.shape[:-1] + explicit_posterior_shape
        )

        # posterior.rsample_from_base_samples require an implicity m dim for MultivariateNormal. Squash last dimension.
        if type(posterior.distribution) is MultivariateNormal:
            return expanded_explicit_base_sample.squeeze(-1)

        return expanded_explicit_base_sample

    # Keeping this as a static function as it useful to child methods.
    @classmethod
    def _explicit_posterior_shape(cls, posterior: GPyTorchPosterior) -> torch.Size:
        """Helper methods to return the posterior shape in format (*b,n,m).

        Different posterior distributions have different number of dimensions:
        - MultivariateNormal has shape (*b,n) (with an implicit m=1)
        - MultitaskMultivariateNormal has shape (*b,n,m)

        This is often cumbersom to work with. Instead we work with the more useful format (*b,n,m).

        Args:
            posterior: posterior to determine the explicit shape of.

        Return:
            torch.Size: for shape (*b,n,m)
        """
        dist_type = type(posterior.distribution)
        if dist_type == MultivariateNormal:
            return posterior.shape() + torch.Size([1])
        if dist_type == MultitaskMultivariateNormal:
            return posterior.shape()

        # We do not expect different distibutions, but the ax/botorch codebase is large, so making this explicity.
        msg = (
            f"Expected posterior.distribution of type MultivariateNormal or MultitaskMultivariateNormal."
            f"Got {dist_type}"
        )
        raise ValueError(msg)


def diagonalize_distribution(posterior: GPyTorchPosterior) -> GPyTorchPosterior:
    """Diagonalize the distribution of the posterior.

    The points in the posterior need to be treated as independent from each other.
    Therefore we want the covariance matrix to be diagonal.

    Args:
        posterior: The posterior to diagonalize.

    Returns:
        The diagonalized posterior.
    """
    mean = posterior.distribution.mean
    # shape is: (*b,n*m diagonal)
    var = posterior.covariance_matrix.diagonal(dim1=-2, dim2=-1)

    diagonal_covar = linear_operator.operators.DiagLinearOperator(var)

    dist_class = type(posterior.distribution)
    if dist_class is MultitaskMultivariateNormal:
        # Need to know if the original covairance was organised in an interleaved way so can instantiate correctly.
        # (see tests\sampling\helpers.py posterior_n3_t2_interleaved for example of interleaved variance).)
        if posterior._interleaved:  # noqa: SLF001
            return GPyTorchPosterior(MultitaskMultivariateNormal(mean, diagonal_covar, interleaved=True))

        return GPyTorchPosterior(MultitaskMultivariateNormal(mean, diagonal_covar, interleaved=False))

    return GPyTorchPosterior(dist_class(mean, diagonal_covar))



import pytest
import torch
from botorch.posteriors.gpytorch import GPyTorchPosterior

from axtreme.sampling import NormalIndependentSampler
from tests.sampling.helpers import (
    posterior_b2_n3_t2,
    posterior_n1_t1,
    posterior_n1_t2,
    posterior_n2_t1,
    posterior_n3_t2,
)


### unit test
@pytest.mark.parametrize(
    "posterior, sample_shape ,expected_base_sample_shape",
    [
        (posterior_n1_t1(), torch.Size([7, 7]), torch.Size([7, 7, 1])),  # MultivariateNormal
        (posterior_n2_t1(), torch.Size([7, 7]), torch.Size([7, 7, 1])),  # MultivariateNormal multiple points
        (posterior_n3_t2(), torch.Size([7, 7]), torch.Size([7, 7, 2])),  # MultitaskMultivariateNormal
        # MultitaskMultivariateNormal with additional batching
        (posterior_b2_n3_t2(), torch.Size([7, 7]), torch.Size([7, 7, 2])),
    ],
)
def test_required_base_sample_shape(
    posterior: GPyTorchPosterior, sample_shape: torch.Size, expected_base_sample_shape: torch.Size
):
    """Check this works for different types of posteriors."""
    actual_shape = NormalIndependentSampler._required_base_sample_shape(sample_shape, posterior)
    assert expected_base_sample_shape == actual_shape


@pytest.mark.parametrize(
    "posterior, expected_base_sample_shape",
    [
        (posterior_n1_t1(), torch.Size([3, 1])),  # MultivariateNormal
        (posterior_n1_t2(), torch.Size([3, 2])),  # MultitaskMultivariateNormal
    ],
)
def test_construct_base_samples_created_with_right_shape(
    posterior: GPyTorchPosterior, expected_base_sample_shape: torch.Size
):
    sampler = NormalIndependentSampler(sample_shape=torch.Size([3]), seed=7)
    sampler._construct_base_samples(posterior)

    assert sampler.base_samples.shape == expected_base_sample_shape


@pytest.mark.parametrize(
    "posterior1, posterior2",
    [
        (posterior_n1_t1(), posterior_n2_t1()),  # MultivariateNormal
        (posterior_n1_t2(), posterior_n3_t2()),  # MultitaskMultivariateNormal
        (posterior_n1_t2(), posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
    ],
)
def test_construct_base_samples_reuse_base_samples_when_possible(
    posterior1: GPyTorchPosterior, posterior2: GPyTorchPosterior
):
    """Reuse base samples if the samples shape and output dim t don't change."""
    sampler = NormalIndependentSampler(sample_shape=torch.Size([3]), seed=7)
    sampler._construct_base_samples(posterior1)
    original_base_samples = id(sampler.base_samples)

    sampler._construct_base_samples(posterior2)
    base_samples_after_repeat = id(sampler.base_samples)

    # The same obect should be used
    assert original_base_samples == base_samples_after_repeat


def test_construct_base_samples_for_posterior_with_new_target_size():
    sampler = NormalIndependentSampler(sample_shape=torch.Size([3]), seed=7)
    sampler._construct_base_samples(posterior_n1_t1())
    base_samples_1 = sampler.base_samples
    sampler._construct_base_samples(posterior_n1_t2())
    base_samples_2 = sampler.base_samples

    assert base_samples_1.shape == torch.Size([3, 1])
    assert base_samples_2.shape == torch.Size([3, 2])


def test_construct_base_samples_for_new_sample_shape():
    """NOTE: don't really expect this usecase, but it is supported."""
    sampler = NormalIndependentSampler(sample_shape=torch.Size([3]), seed=7)
    sampler._construct_base_samples(posterior_n1_t1())
    base_samples_1 = sampler.base_samples
    sampler.sample_shape = torch.Size([4, 4])
    sampler._construct_base_samples(posterior_n1_t1())
    base_samples_2 = sampler.base_samples

    assert base_samples_1.shape == torch.Size([3, 1])
    assert base_samples_2.shape == torch.Size([4, 4, 1])


@pytest.mark.parametrize(
    "posterior, sample_shape, expected_sample_shape",
    [
        (posterior_n1_t1(), torch.Size([1]), torch.Size([1, 1, 1])),  # MultivariateNormal
        (posterior_n1_t1(), torch.Size([3, 2]), torch.Size([3, 2, 1, 1])),  # MultivariateNormal
        (posterior_n1_t2(), torch.Size([3]), torch.Size([3, 1, 2])),  # MultitaskMultivariateNormal
        (posterior_n1_t2(), torch.Size([3, 2]), torch.Size([3, 2, 1, 2])),  # MultitaskMultivariateNormal
        (posterior_n3_t2(), torch.Size([5]), torch.Size([5, 3, 2])),  # MultitaskMultivariateNormal
        (posterior_n3_t2(), torch.Size([5, 2]), torch.Size([5, 2, 3, 2])),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2(), torch.Size([1]), torch.Size([1, 2, 3, 2])),  # MultitaskMultivariateNormal batched
        (posterior_b2_n3_t2(), torch.Size([5, 4]), torch.Size([5, 4, 2, 3, 2])),  # MultitaskMultivariateNormal batched
    ],
)
def test_forward(posterior: GPyTorchPosterior, sample_shape: torch.Size, expected_sample_shape: torch.Size):
    sampler = NormalIndependentSampler(sample_shape=sample_shape)
    samples = sampler(posterior)

    assert samples.shape == expected_sample_shape



# sourcery skip: no-loop-in-tests

# %%
import matplotlib.pyplot as plt
import numpy as np
import pytest
import torch
from scipy.stats import t as scipy_student_t
from torch import Tensor, tensor
from torch.distributions import Distribution, Gumbel, StudentT, Uniform, Weibull

from axtreme.utils.population_estimators import estimate_pdf_value_from_sample, sample_mean_se, sample_median_se


@pytest.mark.non_deterministic
@pytest.mark.parametrize(
    "true_dist",
    [
        # Try a variety of different distributions
        (Weibull(3, 2)),
        (Uniform(3, 5)),
        (Gumbel(3, 2)),
        # See the impact of larger mean
        (Gumbel(30, 2)),
        # see impact of larger variance:
        (Gumbel(30, 6)),
    ],
)
def test_sample_mean_se(true_dist: Distribution, n_samples_per_est: int = 20, n_ests: int = 5000):
    """Checks if the distribution returned by sample_mean_se is reasonable.

    The consifidence bounds returned by this method are frequently used. We want to check that they are well calibrated.
    If we obtain at 95% confidence from a sample, 95% of the time the true mean should fall within these bounds. Higher
    or lower percentage represent miscallibrate.

    NOTE: sample_mean_se is primarily used as a diagnoistic, so we are willing to accept some error in the results.

    Args:
        true_dist: The true ditribution from which to generate samples from. The method has no restriction of the type
          of sample distribution.
    """
    true_mean = true_dist.mean

    samples = true_dist.sample(torch.Size([n_ests, n_samples_per_est]))

    # If the confidence bounds are good, the true mean should fall in the range [0.025, 0.975] 95% of the time.
    # Therefore in our results, we should see q(true_mean) exceeding 0.975 or below 0.025 5% percent of the time.
    quantil_of_true_mean = []
    for sample in samples:
        dist: StudentT = sample_mean_se(sample)
        # The below raises a NotImplementedError, a bit a a hacky fix to switch scipy dist.
        # confidence_region = dist.cdf(1)
        dist_scipy = scipy_student_t(df=int(dist.df), loc=float(dist.loc), scale=float(dist.scale))
        quantil_of_true_mean.append(dist_scipy.cdf(true_mean))

    results = torch.tensor(quantil_of_true_mean)
    percent_outside_of_bounds = torch.logical_or(results > 0.975, results < 0.025).float().mean()

    assert float(percent_outside_of_bounds) == pytest.approx(0.05, abs=0.03)


@pytest.mark.non_deterministic
@pytest.mark.parametrize(
    "true_dist",
    [
        # Try a variety of different distributions
        (Weibull(3, 2)),
        (Uniform(3, 5)),
        (Gumbel(3, 2)),
        # See the impact of larger meanS
        (Gumbel(30, 2)),
        # see impact of larger variance:
        (Gumbel(30, 6)),
    ],
)
def test_sample_median_se(true_dist: Distribution, n_samples_per_est: int = 50, n_ests: int = 5000):
    """Checks if the distribution returned by sample_median_se is reasonable.

    The consifidence bounds returned by this method are frequently used. We want to check that they are well calibrated.
    If we obtain at 95% confidence from a sample, 95% of the time the true mean should fall within these bounds. Higher
    or lower percentage represent miscallibrate.

    NOTE: sample_median_se is primarily used as a diagnoistic, so we are willing to accept some error in the results.

    Args:
        true_dist: The true ditribution from which to generate samples from. The method has no restriction of the type
          of sample distribution.
    """
    true_median = true_dist.icdf(tensor(0.5))

    samples: Tensor = true_dist.sample(torch.Size([n_ests, n_samples_per_est]))

    # If the confidence bounds are good, the true median should fall in the range [0.025, 0.975] 95% of the time.
    # Therefore in our results, we should see q(true_median) exceeding 0.975 or below 0.025 5% percent of the time.
    quantile_of_true_median = []
    for sample in samples:
        dist = sample_median_se(sample)
        quantile_of_true_median.append(dist.cdf(true_median))

    results = torch.tensor(quantile_of_true_median)
    percent_outside_of_bounds = torch.logical_or(results > 0.975, results < 0.025).float().mean()

    assert float(percent_outside_of_bounds) == pytest.approx(0.05, abs=0.05)


@pytest.mark.non_deterministic
@pytest.mark.parametrize(
    "true_dist",
    [
        # Try a variety of different distributions
        (Weibull(3, 2)),
        (Uniform(3, 5)),
        (Gumbel(3, 2)),
        # See the impact of larger meanS
        (Gumbel(30, 2)),
        # see impact of larger variance:
        (Gumbel(30, 6)),
    ],
)
def test_estimate_pdf_value_from_sample_at_median(true_dist: Distribution, n_ests: int = 5000):
    """This tests for degredation in the performance on estimate_pdf_value_from_sample_at_median.

    The bounds required depened on `n_samples_per_est`. `visualise_performance_of_estimate_pdf_value_from_sample` can be
    used to determine them.
    """
    n_samples_per_est = 50

    q: float = 0.5
    _x: Tensor = true_dist.icdf(tensor(q))
    true_pdf_value = true_dist.log_prob(_x).exp()
    x: float = float(_x)

    samples = true_dist.sample(torch.Size([n_ests, n_samples_per_est]))

    pdf_ests = [estimate_pdf_value_from_sample(sample, x) for sample in samples]

    results = torch.tensor(pdf_ests)

    assert float(results.mean()) == pytest.approx(true_pdf_value, rel=0.2)
    coeffeceint_of_variation = results.std() / results.mean()
    assert coeffeceint_of_variation < 0.2


def visualise_performance_of_estimate_pdf_value_from_sample():
    """Show the bias and variance of `estimate_pdf_value_from_sample` an n_samples increases"""

    def test_estimate_pdf_value_from_sample(true_dist: Distribution, n_samples_per_est: int = 50, n_ests: int = 5000):
        q: float = 0.5
        _x: Tensor = true_dist.icdf(tensor(q))
        true_pdf_value = true_dist.log_prob(_x).exp()
        x: float = float(_x)

        samples = true_dist.sample(torch.Size([n_ests, n_samples_per_est]))

        # If the confidence bounds are good, the true mean should fall in the range [0.025, 0.975] 95% of the time.
        # Therefore in our results, we should see q(true_mean) exceeding 0.975 or below 0.025 5% percent of the time.
        pdf_ests = [estimate_pdf_value_from_sample(sample, x) for sample in samples]

        results = torch.tensor(pdf_ests)

        # assert float(results.mean()) == pytest.approx(true_pdf_value, rel=0.01)
        coeffeceint_of_variation = results.std() / results.mean()
        # assert coeffeceint_of_variation < 0.01

        return true_pdf_value, results.mean(), coeffeceint_of_variation

    dists = {
        "Weibull(3, 2)": (Weibull(3, 2)),
        "Uniform(3, 5)": (Uniform(3, 5)),
        "Gumbel(3, 2)": (Gumbel(3, 2)),
        # See the impact of larger meanS
        "Gumbel(30, 2)": (Gumbel(30, 2)),
        # see impact of larger variance:
        "Gumbel(30, 6)": (Gumbel(30, 6)),
    }
    n_samples_list = [11, 22, 44, 88, 176]

    # Placeholder for results
    results: dict[str, dict[str, list[float]]] = {
        name: {"n_samples": [], "true_pdf_value": [], "mean_est": [], "cov": []} for name in dists
    }

    # Run estimation
    for n_samples_per_est in n_samples_list:
        for name, dist in dists.items():
            # Mockup: replace this with your actual function
            true_pdf_value, mean_est, cof = test_estimate_pdf_value_from_sample(dist, n_samples_per_est)

            # Save results
            results[name]["n_samples"].append(n_samples_per_est)
            results[name]["true_pdf_value"].append(true_pdf_value)
            results[name]["mean_est"].append(mean_est)
            results[name]["cov"].append(cof)

    # Plotting results:
    _, axes = plt.subplots(3, 1, figsize=(10, 6 * 3))

    colour_list = ["r", "g", "b", "y", "m", "c", "k"]
    for (name, data), c in zip(results.items(), colour_list, strict=False):
        n_samples = data["n_samples"]
        true_pdf = np.array(data["true_pdf_value"])
        mean_est = np.array(data["mean_est"])
        cov = data["cov"]

        # Plot true_pdf_value and mean_est
        axes[0].plot(n_samples, true_pdf, label=f"{name} (True pdf values)", linestyle="--", c=c)
        axes[0].plot(n_samples, mean_est, label=f"{name} (estimator mean est)", c=c)
        axes[0].set_xlabel("Number of Samples")
        axes[0].set_ylabel("pdf")

        # proportion of estimate
        axes[1].plot(n_samples, mean_est / true_pdf, label=f"{name} mean_est/true", linestyle="--", c=c)
        axes[1].set_xlabel("Number of Samples")
        axes[1].set_ylabel("estimator_bias/ true value")

        # coefficient of variation
        axes[2].plot(n_samples, cov, label=f"{name} cov")
        axes[2].set_xlabel("Number of Samples")
        axes[2].set_ylabel("coeffecint of variation")



import matplotlib.pyplot as plt
import pytest
import torch
from botorch.models import SingleTaskGP

from axtreme.plotting.gp_fit import plot_1d_model
from axtreme.sampling import NormalIndependentSampler, PosteriorSampler, UTSampler


@pytest.mark.integration
@pytest.mark.parametrize(
    "sampler, expected_value",
    [
        (
            NormalIndependentSampler(sample_shape=torch.Size([5]), seed=7),
            torch.tensor(
                [
                    [[0.5811], [0.0562], [-0.9330], [0.0768], [-0.4511]],
                    [[1.2794], [0.8726], [-0.6471], [0.8613], [-0.1794]],
                    [[1.3997], [1.0132], [-0.5979], [0.9965], [-0.1326]],
                    [[-0.1431], [-0.7904], [-1.2295], [-0.7369], [-0.7330]],
                    [[1.9565], [1.6641], [-0.3699], [1.6221], [0.0841]],
                ]
            ),
        ),
        (
            UTSampler(),
            torch.tensor(
                [
                    [[0.6910], [0.1847], [-0.8881], [0.2002], [-0.4084]],
                    [[1.9874], [1.7002], [-0.3573], [1.6568], [0.0962]],
                    [[-0.6054], [-1.3309], [-1.4188], [-1.2564], [-0.9129]],
                ]
            ),
        ),
    ],
)
def test_sampler_integration(
    sampler: PosteriorSampler,
    expected_value: torch.Tensor,
    model_singletaskgp_d1_t1: SingleTaskGP,
    *,
    visual_inspect: bool = False,
):
    """One of the most useful ways to inspect the performance of a sampler is to visually inspect it.

    This test confirms the output matches the visual inspecation (expected_value come from the viual inspection.

    Args:
        sampler: _description_
        expected_value: _description_
        model_singletaskgp_d1_t1: _description_
        visual_inspect: _description_. Defaults to False.

    Todo:
        - A posterior could be passed directly here instead of the model. model has been use so different vidualisations
          can easily be made.
    """
    model = model_singletaskgp_d1_t1
    x = torch.linspace(0, 1, 5).reshape(-1, 1)

    with torch.no_grad():
        posterior = model.posterior(x)

    samples = sampler(posterior)

    if visual_inspect:
        # For better visual inspection it is more useful to increase the number of points in x
        ax = plot_1d_model(model)
        for sample in samples:
            _ = ax.plot(x.flatten(), sample, color="grey")
        plt.pause(5)

    torch.testing.assert_close(samples, expected_value, rtol=1e-3, atol=1e-4)



import pytest
import torch
from botorch.posteriors.gpytorch import GPyTorchPosterior

from axtreme.sampling import MeanSampler
from tests.sampling.helpers import (
    posterior_b2_n3_t2,
    posterior_n1_t1,
    posterior_n1_t2,
    posterior_n2_t1,
    posterior_n3_t2,
)


# While this is a simple test, opted to paramterise so that the output dimension is explicit
@pytest.mark.parametrize(
    "posterior,expected_shape",
    [
        (posterior_n1_t1(), torch.Size([1, 1, 1])),
        (posterior_n2_t1(), torch.Size([1, 2, 1])),
        (posterior_n1_t2(), torch.Size([1, 1, 2])),
        (posterior_n3_t2(), torch.Size([1, 3, 2])),
        (posterior_b2_n3_t2(), torch.Size([1, 2, 3, 2])),
    ],
)
def test_forward_takes_one_sample(posterior: GPyTorchPosterior, expected_shape: torch.Size):
    """Uses posterior (parameterised fixture) to test all types of posteriors because expected output is simple."""
    sampler = MeanSampler()
    samples = sampler(posterior)

    # MeanSampler always samples exactly one samples (just collect the mean)
    # Is adds a dimension infront so that it is consisten with the other samplers
    assert samples.shape == expected_shape



"""Plotting helper tailored for analysingQoIJobResults."""

import copy
from collections.abc import Callable

import matplotlib.pyplot as plt
import pandas as pd
import torch
from botorch.models import PosteriorMeanModel, SingleTaskGP
from botorch.sampling.index_sampler import IndexSampler
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas.core.groupby.generic import DataFrameGroupBy

from axtreme.qoi.gp_bruteforce import GPBruteForce


def plot_distribution(
    df: pd.DataFrame,
    ax: Axes,
    n_hists: int = 3,
    col_name: str = "samples",
    brute_force: float | None = None,
) -> None:
    """Helperfor plotting histograms (on the given ax) of dataframe cells containing lists.

    Designed for use with the 'samples' column of a QoiJobResults dataframe.

    Args:
        df: A dataframe.
        ax: The axis to plot on.
        n_hists: The number of cells of column `col_name` to plot.
        col_name: The column of the df containing lists.
        brute_force: Represents the true value (e.g mean). Plots a vertical line if provided.
    """
    samples_list = df.loc[:, col_name].to_numpy()
    samples_list = samples_list[:n_hists]

    for samples_i in samples_list:
        _ = ax.hist(samples_i, density=True, alpha=0.3, bins=len(samples_i) // 5 + 1)
    _ = ax.set_xlabel("QOI value")
    _ = ax.set_title("QOi estimator distibutions")

    if brute_force:
        _ = ax.axvline(brute_force, c="black", label=f"Brute force ({brute_force:.2f})")

    _ = ax.legend()


def plot_col_histogram(df: pd.DataFrame, ax: Axes, col_name: str = "mean", brute_force: float | None = None) -> None:
    """Helper which creates a histogram (on the given ax) based on a column of the df.

    Designed for use with the 'mean' or 'var' column of a QoiJobResults dataframe.

    Args:
        df: A dataframe.
        ax: The axis to plot on.
        col_name: The column of the df containing lists.
        brute_force: Represents the true value (e.g mean). Plots a vertical line if provided.
    """
    values = df.loc[:, col_name].to_numpy()
    _ = ax.hist(values, label=f"{col_name} of qoi runs", density=True, bins=30)
    _ = ax.set_title(
        f"{col_name} of each qoi run\n mean of dist {values.mean():.3f}."
        f" std of dist {values.std():.3f}, C.o.V {values.std()/ values.mean():.3f}"
    )
    _ = ax.set_ylabel("density")
    _ = ax.set_xlabel("QOI value")

    if brute_force:
        _ = ax.axvline(brute_force, c="black", label=f"Brute force ({brute_force:.2f})")

    _ = ax.legend()


def plot_groups(
    # Error if try to parmaterise this as a generic
    df_grouped: DataFrameGroupBy,  ## pyright: ignore[reportUnknownParameterType,reportMissingTypeArgument]
    plotting_funcs: list[Callable[[pd.DataFrame, Axes], None]],
) -> Figure:
    """Takes a grouped dataframe, and generates a row of plots for each group, using plotting_funcs.

    Args:
        df_grouped: The groupby object to plot
        plotting_funcs: list of plots to be generated for each group. See plot_col_histogram for an example of a
          plotting function.
    """
    group_names = df_grouped.groups
    n_groups = len(group_names)
    n_cols = len(plotting_funcs)
    fig, axes = plt.subplots(nrows=n_groups, ncols=n_cols, figsize=(6 * n_cols, 4 * n_groups), sharex="col")

    if n_groups == 1:
        # reshape so can be handled in the same manner as 2d
        axes = axes.reshape(1, n_cols)

    for i, (_, group_data) in enumerate(df_grouped):
        row_axes = axes[i]
        for plot_func, ax in zip(plotting_funcs, row_axes, strict=True):
            plot_func(group_data, ax)

    # Add group labels to each row
    for ax, row in zip(axes[:, 0], group_names, strict=True):
        ax.annotate(
            row,
            xy=(0, 0.5),
            xytext=(-ax.yaxis.labelpad - 5, 0),
            xycoords=ax.yaxis.label,
            textcoords="offset points",
            size="large",
            ha="right",
            va="center",
        )

    fig.tight_layout()
    # tight_layout doesn't take these labels into account. Manually tweak.
    fig.subplots_adjust(left=0.15)

    return fig


def qoi_ignoring_gp_uncertainty(qoi: GPBruteForce, model: SingleTaskGP) -> torch.Tensor:
    """Helper to run a QoI with a model, ignoring uncertainty in the model (e.g using the posterior mean).

    Args:
        qoi: The QoI estimator to use
        model: The model to use

    Return:
        the estimates made by the QoI using only the posterior mean of the model.
    """
    # copy the QoI as we will need to change some internal settings
    new_qoi = copy.deepcopy(qoi)

    # make a deterministic model
    det_model = PosteriorMeanModel(model)

    # We need to change the sampler to be compatible with a deterministic model

    if not hasattr(new_qoi.posterior_sampler, "sample_shape"):
        msg = (
            "This function only works with samplers that have a sample_shape attribute."
            " e.g Like those inheriting from botorch MCSampler."
        )
        raise NotImplementedError(msg)

    sample_shape = new_qoi.posterior_sampler.sample_shape  # pyright: ignore[reportAttributeAccessIssue]
    sampler = IndexSampler(sample_shape)
    # We need to do the following work around to make IndexSampler compatible with TransformedPosterior. IndexSampler is
    # only designed to work with EnsembelPosterior, but when using outcome transforms on the model, the
    # EnsembelPosterior gets wrapped in a TransformedPosterior. This only causes issues when base samples are created
    # internally, so we create them here instead.
    # TODO(sw 2024-11-22): raise a ticket with botorch to fix this issue?
    sampler.register_buffer("base_samples", torch.zeros(sample_shape, dtype=torch.int))

    new_qoi.posterior_sampler = sampler

    return new_qoi(model=det_model)



# %%
import copy

import pytest
import torch
from botorch.posteriors.gpytorch import GPyTorchPosterior
from gpytorch.distributions.multivariate_normal import MultivariateNormal

from axtreme.sampling.independent_sampler import IndependentMCSampler, diagonalize_distribution
from tests.sampling.helpers import (
    posterior_b2_n3_t2,
    posterior_n1_t1,
    posterior_n2_t1,
    posterior_n3_t2,
    posterior_n3_t2_interleaved,
)


@pytest.mark.parametrize(
    "posterior",
    [
        (posterior_n1_t1()),  # test single input
        (posterior_n2_t1()),  # MultivariateNormal
        (posterior_n3_t2()),  # MultitaskMultivariateNormal
        (posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
        (posterior_n3_t2_interleaved()),  # MultitaskMultivariateNormal interleaved
    ],
)
def test_diagonalize_distribution_remove_covariance(posterior: GPyTorchPosterior):
    ### set up the test
    # make a new posterior with covariance, by adding something to the off diagonal.
    # This is done by adding to the whole covariance matrix, and then setting the diagonal back to their original values
    posterior_with_covar = copy.deepcopy(posterior)
    _ = torch.add(posterior_with_covar.covariance_matrix, 0.01, out=posterior_with_covar.covariance_matrix)
    # reset the diagonals to their original
    n_points = posterior_with_covar.covariance_matrix.shape[-1]
    posterior_with_covar.covariance_matrix[..., range(n_points), range(n_points)] = (
        posterior.covariance_matrix.diagonal(dim1=-2, dim2=-1)
    )

    ### Run the code
    diag_posterior = diagonalize_distribution(posterior_with_covar)

    ### Check the output
    # There is not inbuilt equality for distibutions, so check the key components.
    torch.testing.assert_close(posterior.mean, diag_posterior.mean)
    torch.testing.assert_close(posterior.covariance_matrix, diag_posterior.covariance_matrix)
    assert posterior.base_sample_shape == diag_posterior.base_sample_shape
    # NOTE: we accept .lazy_covariance_matrix being different type (DiagLinearOperator, rather than DenseLinearOperator)

    # Check the attributes from the parent class pytorch.Distibutions.
    # Recommends using _extended_shape instead of event_shape
    assert posterior._extended_shape() == diag_posterior._extended_shape()
    assert posterior.batch_shape == diag_posterior.batch_shape


@pytest.mark.parametrize(
    "base_samples, posterior",
    [
        # Try different distibutions
        (torch.ones([1, 1]), posterior_n2_t1()),  # MultivariateNormal minimal example of broadcasting to 2 points
        (torch.ones([1, 2]), posterior_n3_t2()),  # MultitaskMultivariateNormal
        (torch.ones([1, 2]), posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
        # try different batch shapes
        (torch.ones([7, 7, 1]), posterior_n2_t1()),  # MultivariateNormal minimal example of broadcasting to 2 points
        (torch.ones([7, 7, 2]), posterior_n3_t2()),  # MultitaskMultivariateNormal
        (torch.ones([7, 7, 2]), posterior_b2_n3_t2()),  # MultitaskMultivariateNormal batched
    ],
)
def test_extend_base_samples_correct_output_shape(base_samples: torch.Tensor, posterior: GPyTorchPosterior):
    expanded_samples = IndependentMCSampler._extend_base_samples(base_samples, posterior)
    assert expanded_samples.shape == base_samples.shape[:-1] + posterior.shape()


def test_extend_base_samples_check_base_sample_sucessfully_duplicated_MultivariateNormal():
    """Make sure the base sample is applied to each x."""
    base_sample = torch.arange(4).reshape(2, 2, 1)
    posterior = posterior_n2_t1()
    expanded_samples = IndependentMCSampler._extend_base_samples(base_sample, posterior)

    # expanded_samples is (samples_shape, *b,n,t).
    # Check each sample has been spread the posterior points in the corresponding index
    assert (expanded_samples[0][0] == base_sample[0][0]).all()
    assert (expanded_samples[0][1] == base_sample[0][1]).all()
    assert (expanded_samples[1][0] == base_sample[1][0]).all()
    assert (expanded_samples[1][1] == base_sample[1][1]).all()


def test_extend_base_samples_check_base_sample_sucessfully_duplicated_MultitaskMultivariateNormal_batched():
    """Make sure the base sample is applied to each x."""
    base_sample = torch.arange(4).reshape(2, 2)
    posterior = posterior_b2_n3_t2()
    expanded_samples = IndependentMCSampler._extend_base_samples(base_sample, posterior)

    # expanded_samples is (samples_shape, *b,n,t).
    # Check each sample has been spread the posterior points in the corresponding index

    # expanded_samples[0]: (2,3,2)  base_sample[0]: (2)
    # check that all the batches b=2, and all points n=3 has the targets = basesample[0]
    assert (expanded_samples[0] == base_sample[0]).all()
    assert (expanded_samples[0] == base_sample[0]).all()


@pytest.mark.integration
def test_forward_same_samples_used_at_different_n_points_MultivariateNormal():
    r"""Check that the same base sample gets used independantly at each x.

    We override _construct_base_samples' to always reutn the value z = [[1]]. If we then run iwth samples_shape = 1,
    we apply this sample to each of the entries in our posterior. Because independant samplers don't consider the
    covariance between points, we can check each value independantly. Each one should follow this rule:

        $sample = \mu + std * z$   (where z is the base_sample)

    Here we check that using the sampler produce the same result as this.
    """
    manual_base_shape = torch.tensor([[1]])

    class MockIndependentMCSampler(IndependentMCSampler):
        def _construct_base_samples(self, posterior: GPyTorchPosterior) -> None:  # noqa: ARG002
            self.register_buffer("base_samples", manual_base_shape)

    sampler = MockIndependentMCSampler(sample_shape=torch.Size([1]))

    # set up a posterior where it will be be easy to determine if the same samples was used.
    mean = torch.tensor([1, 0.5])
    covar = torch.tensor([[1, 0.1], [0.1, 0.81]])
    posterior = GPyTorchPosterior(MultivariateNormal(mean=mean, covariance_matrix=covar))
    # calucated the independant result for each point
    expect_value_1 = mean[0] + covar[0][0] ** 0.5 * manual_base_shape[0][0]  # should be 1 + 1 = 2
    expect_value_2 = mean[1] + covar[1][1] ** 0.5 * manual_base_shape[0][0]  # should be .5 + .9= 1.4

    # shape (samples_shape = 1,n = 2, m = 1)
    samples = sampler(posterior)

    assert samples[0][0] == expect_value_1
    assert samples[0][1] == expect_value_2


# NOTE: `forward` is checked more extensively in the concerete implementations NoramlIndependentSampler.



"""Helper for working with scipy distibutions."""

import inspect

import numpy as np
from numpy.typing import ArrayLike
from scipy.stats import gumbel_r, rv_continuous
from statsmodels.base.model import GenericLikelihoodModel


# How many params do we need to fit, and what are the names so we can identify them
def distribution_parameter_names_from_scipy(dist: rv_continuous) -> list[str]:
    """Collects the parameters required for a scipy rv_continous distibution.

    Withing axtreme this is used to determine what a GP should predict.

    Args:
        dist: The distibution to retrieve the parameters for.

    Returns:
        The distibutiion parameter names, IN THE ORDER they should be passed to calls to this distibution class.
        For example:

          - gumbel_r.pdf(x, loc=0, scale=1) -> order returned ['loc','scale']
          - weibull_max.pdf(x, c, loc=0, scale=1) -> order returned ['c','loc','scale']

    """
    # All rv_continous use/support 'loc' and 'scale' as detailed here: scipy.stats.rv_continuous
    parameter_names = ["loc", "scale"]
    # Collect any addition parameters the distibutuion defines.
    # This is based on how scipy itself finds the required parameters here:
    # scipy.stats.rv_continuous.__init__ in inside `_construct_argparser`
    # typically accessing priviate object is an anti-pattern, but this how scipy itself determines this
    # and ._pdf is actually user define when making new rv_contious objects
    pdf_params = list(inspect.signature(dist._pdf).parameters.keys())  # noqa: SLF001

    # Drop the first entry of pdf_params as this is always x
    # NOTE: the parameter order is important
    # This is the same order the the dist expects the data to be passed into the function signatures
    dist_params = pdf_params[1:] + parameter_names

    return dist_params


def fit_dist_with_uncertainty(
    data: ArrayLike,
    dist: rv_continuous,
) -> tuple[np.ndarray[tuple[int,], np.dtype[np.float64]], np.ndarray[tuple[int, int], np.dtype[np.float64]]]:
    """Fit a distibution, returning the parameters estimate and Gaussian uncertainty of the fit.

    Results of fitting with Maximum Likelihood give a normal distibution where

    - mean:parameter estimate
    - covariance matrix: Uncertainty in estimate (Inverse Fischer infomation matrix)

    Note:
        Parameters are returned in tha same order they appear in the fucntion calls of this call.
        For example, weibull_max.pdf(x, c, loc=0, scale=1) -> order returned ['c','loc','scale']
        This order can be extracted with the helper function `distribution_parameter_names_from_scipy`

    Args:
        data (ArrayLike): 1 dimensional set of data to fit a distibution to.
        dist (rv_continuous): the distribution to be fitted.

    Returns:
        feature mean: (n) parameter mean estimates
        covariance: (n,n) covariance matrix of the uncertainty in parameters fitted

    Todo:
        - Some distributions have convergence issues. Currently only allow gumble_r.
        - Optimisation method needs thorough testing
    """
    if type(dist) is not type(gumbel_r):
        msg = (
            "Currently only support gumbel, as the optimisation does not always converge with other distributions."
            " If using other distibution will need to perform suffecient testing then add to the allowed types in"
            " this method"
        )
        raise NotImplementedError(msg)

    # Has to return the logprob of dataset given a set of params
    # Typically this is implementented on the object which has the data stored internally
    # Using closure to conform to the function signature and still have access to the data
    def loglike(params: ArrayLike) -> float:
        return np.sum(dist.logpdf(data, *params))

    # GenericLikelihoodModel documentation example shows the data should also be passed here
    # (in addition to loglike function())
    model = GenericLikelihoodModel(endog=data, loglike=loglike)

    # TODO(sw): This needs testing, works to generically pass good starting points, BUT
    # Need to be careful, exact estimates have made other optimisers fail
    fitted_params = dist.fit(data)
    result = model.fit(start_params=fitted_params, disp=False)

    return result.params, result.cov_params()



"""Helper functions for creating vaid simulators."""

import inspect
from collections.abc import Callable

import numpy as np

from axtreme.simulator.base import Simulator


def simulator_from_func(
    func: Callable[
        [np.ndarray[tuple[int, int], np.dtype[np.float64]]], np.ndarray[tuple[int, int], np.dtype[np.float64]]
    ],
) -> Simulator:
    """Adds `n_simulations_per_point` functionality to a a function that takes x inputs, and return simulation values.

    Args:
        func: function which accepts (n,d) x input, and returns (n, m) y output (one response per x point).

    Returns:
        Function conforming to Simultator interface. Output will always be 3 dimensional
        (n_points, n_simulations_per_point, n_targets).
    """

    def simulator(
        x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: int = 1
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        # repeat x to get num_simulations_per_statebounds
        x_expanded = np.repeat(x, n_simulations_per_point, axis=0)

        # evaluate the function
        y = func(x_expanded)

        # reshape y to desired shape (n, n_simulations_per_point, d)
        return y.reshape(x.shape[0], n_simulations_per_point, -1)

    return simulator


def is_valid_simulator(instance: object, *, verbose: bool = False) -> bool:  # noqa: C901, PLR0911
    """Checks that `instance` conforms to Simulator protocol AND signature definition.

    This test only works if the instance has type definitions.

    Note:
        By default, `isinstance` with `@runtime_checkable` "will check only the presence of the required methods not
        their type signatures!". See `@runtime_checkable` docs for details.

    Args:
        instance: will check if this is a valid Simulator.
        verbose: If true will provide information regarding why it does not conform.

    Returns:
        bool: True if methods and signatures match

    Note:
        - Briefly looked at override the __instancecheck__ so isinstance would acheive this result
        - Apprears __instancecheck__ doesn't get called, would need to investigate why to proceed
        - Adding this behaviour to protocol would cause it to depart from standard protocol behaviour
    """
    if not isinstance(instance, Simulator):
        if verbose:
            print("Instance does not have the same methods as `Simulator`") if verbose else None  # noqa: T201
        return False

    # Specific check regarding which parameters need the same name, defaults, and type annotations.
    # For functions can inspect them directly, for classes need to inspect the __call__ method.
    instance_sig = inspect.signature(instance) if inspect.isfunction(instance) else inspect.signature(instance.__call__)

    sim_sig = inspect.signature(Simulator.__call__)

    # check everything in the simulator signature is in the instance signature
    for sim_param, sim_info in sim_sig.parameters.items():
        if sim_param != "self":
            # Check param is available in instance
            if sim_param not in instance_sig.parameters:
                print(f"Input is missing parameter '{sim_param}'") if verbose else None  # noqa: T201
                return False
            # check the type is correct:
            if sim_info.annotation != instance_sig.parameters[sim_param].annotation:
                print(f"Parameter {sim_param} has incorrect type.") if verbose else None  # noqa: T201
                return False
            # If a default is provided as part of sim, instance should have one too
            if (sim_info.default is not inspect.Parameter.empty) and (
                instance_sig.parameters[sim_param].default is inspect.Parameter.empty
            ):
                print(f"Parameter {sim_param} is missing  a default") if verbose else None  # noqa: T201
                return False

    # check any extras in the instance signature has a default.
    for instance_param, instance_info in instance_sig.parameters.items():
        if instance_param != "self" and instance_param not in sim_sig.parameters:  # noqa: SIM102
            if instance_info.default is inspect.Parameter.empty:
                print(  # noqa: T201
                    f"{instance_param} is an additional parameter to Simulator interface, and has no default value."
                ) if verbose else None
                return False

    if instance_sig.return_annotation != sim_sig.return_annotation:
        if verbose:
            print(  # noqa: T201
                f"Parameter return types are different: instance: {instance_sig.return_annotation}"
                f" Expected: {sim_sig.return_annotation}"
            )
        return False

    return True



"""A copy of pytorch's RandomSampler that uses the same random sample for each iteration.

This is basically an identical copy, except the Generator has been swapped for a seed, which creates the same generator
each time the samples are iterated through (e.g the DataLoader is run to completion).
"""

from collections.abc import Iterator, Sized

import torch
from torch.utils.data import Sampler


class FixedRandomSampler(Sampler[int]):
    r"""Samples elements randomly.

    This sampler differs from `torch`'s `RandomSampler` as it will return the same random sample each time it is
    iterated in full. If without replacement, then sample from a shuffled dataset. If with replacement, then user can
    specify :attr:`num_samples` to draw.
    """

    data_source: Sized
    replacement: bool

    def __init__(
        self, data_source: Sized, num_samples: int | None = None, seed: int | None = None, *, replacement: bool = False
    ) -> None:
        """Initalise the sampler.

        Args:
            data_source: dataset to sample from
            replacement: samples are drawn on-demand with replacement if ``True``, default=``False``
            num_samples: number of samples to draw, default=`len(dataset)`.
            seed: seed for the random number generator, if `None` one will be allocated randomly.
        """
        self.data_source = data_source
        self.replacement = replacement
        self._num_samples = num_samples
        self.seed = seed or int(torch.empty((), dtype=torch.int64).random_().item())

        if not isinstance(self.replacement, bool):  # pyright: ignore[reportUnnecessaryIsInstance]
            raise TypeError(f"replacement should be a boolean value, but got replacement={self.replacement}")

        if not isinstance(self.num_samples, int) or self.num_samples <= 0:  # pyright: ignore[reportUnnecessaryIsInstance]
            raise ValueError(f"num_samples should be a positive integer value, but got num_samples={self.num_samples}")

    @property
    def num_samples(self) -> int:  # noqa: D102
        # dataset size might change at runtime
        if self._num_samples is None:
            return len(self.data_source)
        return self._num_samples

    def __iter__(self) -> Iterator[int]:  # noqa: D105
        generator = torch.Generator()
        _ = generator.manual_seed(self.seed)

        n = len(self.data_source)

        if self.replacement:
            for _ in range(self.num_samples // 32):
                yield from torch.randint(high=n, size=(32,), dtype=torch.int64, generator=generator).tolist()
            yield from torch.randint(
                high=n, size=(self.num_samples % 32,), dtype=torch.int64, generator=generator
            ).tolist()
        else:
            for _ in range(self.num_samples // n):
                yield from torch.randperm(n, generator=generator).tolist()
            yield from torch.randperm(n, generator=generator).tolist()[: self.num_samples % n]

    def __len__(self) -> int:  # noqa: D105
        return self.num_samples



"""System test of qoi.GPBruteForce by running increasingly realistic version of the QoI and checking consistency.

`test_qoi_brute_force_system_test` is the function which orchestrates the test. The following function orchestrate the
individual steps, and contain additional detail regarding the motivation for the test.
- ground_truth_estimate
- qoi_no_gp
- qoi_gp_deterministic
- qoi_gp_low_uncertainty
- qoi_gp_high_uncertainty

This script is designed to be run interactively as well as though pytest.
"""

# pyright: reportUnnecessaryTypeIgnoreComment=false

# %%
import json
import time
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
import pytest
import torch
from ax import Models
from botorch.models import SingleTaskGP
from botorch.models.deterministic import GenericDeterministicModel, PosteriorMeanModel
from botorch.sampling.index_sampler import IndexSampler
from matplotlib.axes import Axes
from numpy.typing import NDArray
from torch.utils.data import DataLoader, Dataset, RandomSampler
from tqdm import tqdm

from axtreme.data import BatchInvariantSampler2d, MinimalDataset
from axtreme.eval import utils
from axtreme.eval.qoi_helpers import plot_col_histogram, plot_distribution, plot_groups
from axtreme.eval.qoi_job import QoIJob, QoIJobResult
from axtreme.experiment import add_sobol_points_to_experiment
from axtreme.qoi import GPBruteForce
from axtreme.sampling import NormalIndependentSampler
from axtreme.utils import population_estimators, transforms

# If running interactively, this is required to find examples
if __name__ == "__main__":
    import sys

    root_dir = Path(__file__).parent.parent.parent
    sys.path.append(str(root_dir))

from examples.demo2d.problem import brute_force, env_data, simulator
from examples.demo2d.problem.experiment import make_experiment

# %%
# TODO(sw 2024-11-19): To fit better with pytest this should be global
N_ENV_SAMPLES_PER_PERIOD = 1000

_: Any


# %% Prep plotting function to include brute force
def plot_best_guess(df: pd.DataFrame, ax: Axes, brute_force: float | None = None) -> None:
    """plot_col_histogram with SE added around the mean.

    This function conforms to the interface in axtreme.eval.qoi_plots
    """
    plot_col_histogram(df, ax, brute_force=brute_force)

    # Add the se plot
    ax_twin = ax.twinx()
    samples = torch.tensor(df.loc[:, "mean"].to_numpy())
    sample_mean_se_dist = population_estimators.sample_mean_se(samples)
    _ = population_estimators.plot_dist(sample_mean_se_dist, ax=ax_twin, c="red", label="dist mean 99.7% conf interval")
    _ = ax_twin.legend()


def best_guess_mean_vs_true_value(df: pd.DataFrame, true_value: float) -> torch.Tensor:
    """Helper to return the z score of the true value under the sample mean distribution.

    Args:
        df: Dataframe expected to have column 'mean'
        true_value: the population mean
    """
    samples = torch.tensor(df.loc[:, "mean"].to_numpy())
    sample_mean_se_dist = population_estimators.sample_mean_se(samples)

    return (true_value - sample_mean_se_dist.mean) / sample_mean_se_dist.stddev


def collect_statistics(df: pd.DataFrame, true_value: float) -> dict[str, float]:
    """Helper to return a dictionary of statistics for a dataframe.

    Args:
        df: expected to contain columns 'mean', 'var' and 'samples'.
        true_value: the true value of the population mean.

    Returns:
        Dictionary with the following format:

        {best_guess_z: z score of the true value under the sample mean distribution,
        best_guess_mean: mean of the sample mean distribution,
        best_guess_std: standard deviation of the sample mean distribution,
        var_mean: mean of the variance distribution,
        var_std: standard deviation of the variance distribution}
    """
    return {
        "best_guess_z": float(best_guess_mean_vs_true_value(df, true_value=true_value)),
        "best_guess_mean": df["mean"].mean(),
        "best_guess_std": df["mean"].std(),
        "var_mean": df["var"].mean(),
        "var_std": df["var"].std(),
    }


def get_id() -> str:
    """Produces and id which can be used to label related objects when saving."""
    import git

    repo = git.Repo(search_parent_directories=True)
    sha = repo.head.object.hexsha

    return datetime.now().strftime("%y_%m_%d-%H_%M_%S") + f"__commit_{sha}"  # noqa: DTZ005


@pytest.fixture(scope="module")
def output_dir():
    output_dir = Path(__file__).parent / "results" / get_id()
    output_dir.mkdir(parents=True)
    return output_dir


# To get more information run with `uv run pytest -m system -s``
@pytest.mark.system
@pytest.mark.non_deterministic
def test_qoi_brute_force_system_test(  # noqa: C901, PLR0913, PLR0912, PLR0915
    output_dir: None | Path,
    n_periods: int = 101,
    n_posterior_samples: int = 50,
    n_qoi_runs: int = 50,
    jobs_input_file: None | Path | pd.DataFrame = None,
    error_tol_scaling: float = 1,
    *,
    show_plots: bool = False,
    run_tests: bool = True,
    return_statistics: bool = False,
) -> dict[str, dict[str, float]] | None:
    """System test of QOI_brute_force by running increasingly realistic version of the QoI and checking consistency.

    Args:
        n_periods: number of periods to use. each produces `n_qoi_runs` estimates of the ERD.
        n_posterior_samples: number of posterior samples to take. Each produces an estimate of the QoI.
        n_qoi_runs: Number of times to run the QoIEstimator (a single run produces n_posterior_sampler estimates)
        jobs_input_file: path to json file contain QoIJobResult files.
            - if provided, will be used for testing (This is useful for analysis)
            - if None, jobs will be created and run, and then testing will occure.
        output_dir: directory to save plots to. If none, will not save output.
        error_tol_scaling: The error allowed in assertions is multiplied by this number.
        show_plots: If True plots will be shown.
        run_tests: If True will run assertions on the statistics produced by each qoi.
        return_statistics: If ture the function will return the statistic calculated.

    Details:
    Testing undertakes the following steps:
    - Ground truth estimate: use underlying function directly to make a QoI estimate.
        - Purpose: Gives an estimate of the confidence that can be achieved with a given number of samples. We then
        compare future results to this.
    - QoI with true underlying: use the QoI methodology with the true underling.
        - Purpose: confirm that the QoI implementation (excluding the GP) is correct (matches ground truth).
        - Purpose: Check setting for the number of periods used.
    - QoI with no variance GP: Using a GP with many points, remove uncertainty and calculate QoI.
        - Purpose: Check the quality of fit of the posterior mean. Introduce GP without posterior effects.
    - QoI with minimal variance GP: Using a GP with many points (and therefor very small uncertainty).
        - Purpose: Confirm produces the same results as above after posterior is introduced.
    - QoI with uncertain GP: Use a gp with only a few training points.
        - Purpose: Explore how posterior samples performs in these conditions.
        - Expectations: Higher variance withing a QoI output, potentially introduces bias

    Using this script to test a new QoIEstimator:
    - Start by producing a ground truth: This give you:
        - best guess (and uncertainty in this should expect)
        - variance in the QoI dist that should be expected.
    - Use the QoI using true underlying: This gives you:
        - Confirmation that with a perfect knowledge of the underling function the ground truth can be reporduced
    - Introduce the GP:
        - check using well trained GP it is possible to reproduce the ground truth

    NOTE: the bounds set on the test have been identified empirically. They should only fail 1% off the time. The method
    for obtaining these bounds is at the end of this file.

    Return:
    Nested dictionary.

        {"<QoI_name>": collect_statistics(df)}
    """
    #### Problem parameters
    """Note:
    It is useful to compare models with the same 'budget' of estimates. The budget is primarily a measure of how many
    samples are drawn from the surrogate. e.g in the ground truth the budget for a single QoI is:

    budget = n_periods * N_ENV_SAMPLES_PER_PERIOD * n_posterior_samples

    Use comparable budgets allows apples-to-apples comparison of the confidence and variance of different methods.
    """
    # bookkeeping parameters
    jobs_output_file = output_dir / "qoi_job_results.json" if output_dir else None

    # Problem constants # TODO(sw): come back with a cleaner way to do this
    brute_force_qoi: float = float(
        brute_force.collect_or_calculate_results(period_length=N_ENV_SAMPLES_PER_PERIOD, num_estimates=300_000).median()
    )

    _data = env_data.collect_data()
    env_dataset: Dataset[NDArray[np.float64]] = MinimalDataset(_data.to_numpy())

    ### Helpers:
    # plotting helpers
    plot_distribution_bf = partial(plot_distribution, brute_force=brute_force_qoi)
    plot_dist_variance = partial(plot_col_histogram, col_name="var")
    plot_best_guess_bf = partial(plot_best_guess, brute_force=brute_force_qoi)

    # If passed existing jobs, these should be used instead.
    # This can be somewhat frail, requires consitent naming between rungs
    if isinstance(jobs_input_file, Path):
        with jobs_input_file.open() as fp:
            data = json.load(fp)
            df_jobs = pd.json_normalize(data, max_level=1)
            df_jobs.columns = df_jobs.columns.str.removeprefix("tags.")
    elif isinstance(jobs_input_file, pd.DataFrame):
        df_jobs = jobs_input_file

    # collect results
    statistics = {}

    ##### Run the differest QoIEstimator
    """Ground truth testing:
    Expections:
    - Should be a very good (unbiased estimate) for the true values
    """
    start_time = time.time()
    if jobs_input_file is not None:
        df_ground_truth = df_jobs[df_jobs["name"] == "ground_truth"]
    else:
        df_ground_truth = ground_truth_estimate(n_periods, n_posterior_samples, n_qoi_runs, jobs_output_file)

    statistics["ground_truth"] = collect_statistics(df_ground_truth, brute_force_qoi)

    # Plot results
    fig_ground_truth = plot_groups(
        df_ground_truth.groupby("name"), [plot_distribution_bf, plot_best_guess_bf, plot_dist_variance]
    )
    fig_ground_truth.savefig(str(output_dir / "ground_truth.png")) if output_dir else None
    fig_ground_truth.show() if show_plots else None

    ## Ground truth tests:
    if run_tests:
        stats_ground_truth = statistics["ground_truth"]
        assert abs(stats_ground_truth["best_guess_z"]) < 4 * error_tol_scaling
    print(f"Ground truth {(time.time()-start_time)//60:.0f}:{time.time()-start_time:.2f}")  # noqa: T201
    ##### Qoi_no_gp
    """QoI_no_gp testing
    Expections:
    - Should be a very good (unbiased estimate) for the true values.
    - best guess variance should be very similar
    - variance distribution should be similar
    """
    start_time = time.time()
    if jobs_input_file is not None:
        df_no_gp = df_jobs[df_jobs["name"] == "qoi_no_gp"]
    else:
        df_no_gp = qoi_no_gp(env_dataset, n_periods, n_posterior_samples, n_qoi_runs, jobs_output_file)

    statistics["qoi_no_gp"] = collect_statistics(df_no_gp, brute_force_qoi)

    # plot results
    fig_qoi_no_gp = plot_groups(
        df_no_gp.groupby("name"), [plot_distribution_bf, plot_best_guess_bf, plot_dist_variance]
    )
    fig_qoi_no_gp.savefig(str(output_dir / "qoi_no_gp.png")) if output_dir else None
    fig_qoi_no_gp.show() if show_plots else None

    # Run tests
    if run_tests:
        # fmt: off
        stats_no_gp = statistics["qoi_no_gp"]
        assert abs(stats_no_gp["best_guess_z"]) < 4 * error_tol_scaling
        assert stats_no_gp["best_guess_std"] == pytest.approx(stats_ground_truth["best_guess_std"], rel=0.35 * error_tol_scaling)  # noqa: E501
        assert stats_no_gp["var_mean"] == pytest.approx(stats_ground_truth["var_mean"], rel=0.15 * error_tol_scaling)
        assert stats_no_gp["var_std"] == pytest.approx(stats_ground_truth["var_std"], rel=0.5 * error_tol_scaling)
        # fmt: on

    print(f"QoI_no_gp {(time.time()-start_time)//60:.0f}:{time.time()-start_time:.2f}")  # noqa: T201
    ##### Highly trained GP
    """Deterministic GP
    Expections:
        - Results should be very similar to brute force, potentially we some added bias due to gp fit.
            - Only interested in understanding this bias
    """
    start_time = time.time()
    if jobs_input_file is not None:
        df_gp_deterministic = df_jobs[df_jobs["name"] == "qoi_gp_deterministic"]
    else:
        df_gp_deterministic = qoi_gp_deterministic(
            env_dataset, n_periods, n_posterior_samples, n_qoi_runs, jobs_output_file
        )

    statistics["qoi_gp_deterministic"] = collect_statistics(df_gp_deterministic, brute_force_qoi)

    # plot results
    fig_qoi_gp_deterministic = plot_groups(
        df_gp_deterministic.groupby("name"), [plot_distribution_bf, plot_best_guess_bf, plot_dist_variance]
    )
    fig_qoi_gp_deterministic.savefig(str(output_dir / "qoi_gp_deterministic.png")) if output_dir else None
    fig_qoi_gp_deterministic.show() if show_plots else None

    # Run tests
    if run_tests:
        stats_gp_deterministic = statistics["qoi_gp_deterministic"]
        assert abs(stats_gp_deterministic["best_guess_z"]) < 5 * error_tol_scaling

    print(f"Deterministic GP {(time.time()-start_time)//60:.0f}:{time.time()-start_time:.2f}")  # noqa: T201

    """Gp low variance

    The GP has a very low amount of uncertianty. It is expected to produce a very similar result as above.
    Expected results:
    - produces results very close to the ground truth.
    - Slightly larger variance in best guess (provided gp_deterministic is a good/unbiased estimator)
    """
    start_time = time.time()
    if jobs_input_file is not None:
        df_gp_low_uncertainty = df_jobs[df_jobs["name"] == "qoi_gp_low_uncertainty"]
    else:
        df_gp_low_uncertainty = qoi_gp_low_uncertainty(
            env_dataset, n_periods, n_posterior_samples, n_qoi_runs, jobs_output_file
        )

    statistics["qoi_gp_low_uncertainty"] = collect_statistics(df_gp_low_uncertainty, brute_force_qoi)

    # plot results
    fig_qoi_gp_low_uncertainty = plot_groups(
        df_gp_low_uncertainty.groupby("name"), [plot_distribution_bf, plot_best_guess_bf, plot_dist_variance]
    )
    fig_qoi_gp_low_uncertainty.savefig(str(output_dir / "qoi_gp_low_uncertainty.png")) if output_dir else None
    fig_qoi_gp_low_uncertainty.show() if show_plots else None

    # run tests
    if run_tests:
        stats_low_uncertainty = statistics["qoi_gp_low_uncertainty"]
        # fmt: off
        assert abs(stats_low_uncertainty["best_guess_z"]) < 5 * error_tol_scaling
        assert stats_low_uncertainty["best_guess_std"] == pytest.approx(stats_ground_truth["best_guess_std"], rel=0.6 * error_tol_scaling)  # noqa: E501
        assert stats_low_uncertainty["var_mean"] == pytest.approx(stats_ground_truth["var_mean"], rel=0.15 * error_tol_scaling)  # noqa: E501
        assert stats_low_uncertainty["var_std"] == pytest.approx(stats_ground_truth["var_std"], rel=0.5 * error_tol_scaling)  # noqa: E501
        # fmt: on

    print(f"Gp low variance {(time.time()-start_time)//60:.0f}:{time.time()-start_time:.2f}")  # noqa: T201
    """Gp high variance

    The GP has a larger amount of uncertianty. It is expected to produce moer uncertain results.

    Expected results:
    - There are no strict tests that can be performed here. See method for details.
    """
    start_time = time.time()
    if jobs_input_file is not None:
        df_qoi_gp_high_uncertainty = df_jobs[df_jobs["name"] == "qoi_gp_high_uncertainty"]
    else:
        df_qoi_gp_high_uncertainty = qoi_gp_high_uncertainty(
            env_dataset, n_periods, n_posterior_samples, n_qoi_runs, jobs_output_file
        )

    statistics["qoi_gp_high_uncertainty"] = collect_statistics(df_qoi_gp_high_uncertainty, brute_force_qoi)

    fig_qoi_gp_high_uncertainty = plot_groups(
        df_qoi_gp_high_uncertainty.groupby("name"), [plot_distribution_bf, plot_best_guess_bf, plot_dist_variance]
    )
    fig_qoi_gp_high_uncertainty.savefig(str(output_dir / "qoi_gp_high_uncertainty.png")) if output_dir else None
    fig_qoi_gp_high_uncertainty.show() if show_plots else None
    print(f"Gp low variance {(time.time()-start_time)//60:.0f}:{time.time()-start_time:.2f}")  # noqa: T201

    # TODO(sw 2024-12-9): This is a hacky fix so statistics are easily available when calibrating bounds (see bottom of
    # file), and they are not returned in general (when this is running through pytest). Statistic should probably be
    # saved with the plots as well.
    if return_statistics:
        return statistics

    return None


# %%
def qoi_gp_high_uncertainty(
    env_dataset: Dataset[NDArray[np.float64]],
    n_periods: int,
    n_posterior_samples: int,
    n_qoi_runs: int,
    jobs_output_file: None | Path = None,
) -> pd.DataFrame:
    """Run the QoI with a highly trained model.

    Args:
        env_dataset: Dataset containing the envdata we have access to.
        n_periods: number of periods to use. each produces `n_qoi_runs` estimates of the ERD.
        n_posterior_samples: number of posterior samples to take. Each produces an estimate of the QoI.
        n_qoi_runs: Number of times to run the QoIEstimator (a single run produces n_posterior_sampler estimates)
        jobs_output_file (None | Path): Json file to write raw run results to. If not provided do no write results.

    Return:
        A DataFrame where each row represents a single QoIeEstimator run. Contains columns:
            - mean: The samples mean produced by the QoIeEstimator for that run.
            - var: The samples var produced by the QoIeEstimator for that run.
            - samples: The samples produced by the QoIeEstimator for that run.
            - name: Name of the estimator. This is used to group these results together.

    Details:
        As described in `qoi_gp_low_uncertainty`:
        - Can't check this value against the brute force because imperfect GP will have bias.
        - Can't compared to the PosteriorMean of the GP, as there can be non-linear response.

    Expectations:
        General:
        - Much greater variance within a QoI estimate (each n_est uses a very different underling function)

        Increase the posterior samples will:
        - Reduce the variance in best guesses
        - Wont' reduce the mean of the variance, but should reduce the variance of the variance
    """
    botorch_model = get_trained_gp(n_points=20)

    qoi_gp_high_uncertain_jobs = []
    for i in range(n_qoi_runs):
        qoi_est = GPBruteForce(
            # random dataloader give different env samples for each instance
            env_iterable=get_dataloader(env_dataset, n_periods, batch_size=64),
            # Created new each time to insure not freezing/subsample.
            # because we are not using a deterministic sampler we need this (real) sampler
            posterior_sampler=NormalIndependentSampler(torch.Size([n_posterior_samples])),
            no_grad=True,
        )

        qoi_gp_high_uncertain_jobs.append(
            QoIJob(
                name=f"qoi_gp_high_uncertainty_{i}",
                qoi=qoi_est,
                model=botorch_model,
                tags={"name": "qoi_gp_high_uncertainty"},
            )
        )

    qoi_gp_high_uncertain_results = [job(output_file=jobs_output_file) for job in tqdm(qoi_gp_high_uncertain_jobs)]
    df_jobs = pd.json_normalize([item.to_dict() for item in qoi_gp_high_uncertain_results], max_level=1)
    df_jobs.columns = df_jobs.columns.str.removeprefix("tags.")

    return df_jobs


def qoi_gp_low_uncertainty(
    env_dataset: Dataset[NDArray[np.float64]],
    n_periods: int,
    n_posterior_samples: int,
    n_qoi_runs: int,
    jobs_output_file: None | Path = None,
) -> pd.DataFrame:
    """Run the QoI with a highly trained model.

    Args:
        env_dataset: Dataset containing the envdata we have access to.
        n_periods (int): number of periods to use. each produces `n_qoi_runs` estimates of the ERD.
        n_posterior_samples (int): number of posterior samples to take. Each produces an estimate of the QoI.
        n_qoi_runs (int): Number of times to run the QoIEstimator (a single run produces n_posterior_sampler estimates)
        jobs_output_file (None | Path): Json file to write raw run results to. If not provided do no write results.

    Return:
        A DataFrame where each row represents a single QoIeEstimator run. Contains columns:
            - mean: The samples mean produced by the QoIeEstimator for that run.
            - var: The samples var produced by the QoIeEstimator for that run.
            - samples: The samples produced by the QoIeEstimator for that run.
            - name: Name of the estimator. This is used to group these results together.

    Details:
        The GP has a very low amount of uncertianty. Expected to produce a very similar result as Gp deterministic.

        Note: Uncertinaty in the GP can (correctly) produce bias results. This can occur if the response calcuated from
        the GP is very non-linear/sensitive to the GP. e.g:
        - Sample smaller than the posterior median: result = .1
        - Sample at posterior mean: result = 1
        - Sample larger than the posterior median: result = 10
        When taking the best guess from these results (the mean = 3.7):
        - this is a poor representation of the posterior mean
        - this is a good representation of the response you could get due to uncertainty in the underlying function.

        Expected results:
        - produces results very close to the ground truth.
        - Slightly larger variance is best guess.
    """
    botorch_model = get_trained_gp()

    dataloader = get_dataloader(env_dataset, n_periods, batch_size=64)
    qoi_gp_uncertain_jobs = []
    for i in range(n_qoi_runs):
        qoi_est = GPBruteForce(
            # random dataloader give different env samples for each instance
            env_iterable=dataloader,
            # Created new each time to insure not freezing/subsample.
            # because we are not using a deterministic sampler we need this (real) sampler
            posterior_sampler=NormalIndependentSampler(torch.Size([n_posterior_samples])),
            no_grad=True,
        )

        qoi_gp_uncertain_jobs.append(
            QoIJob(
                name=f"qoi_gp_low_uncertainty_{i}",
                qoi=qoi_est,
                model=botorch_model,
                tags={"name": "qoi_gp_low_uncertainty"},
            )
        )

    qoi_gp_uncertain_results = [job(output_file=jobs_output_file) for job in tqdm(qoi_gp_uncertain_jobs)]
    df_jobs = pd.json_normalize([item.to_dict() for item in qoi_gp_uncertain_results], max_level=1)
    df_jobs.columns = df_jobs.columns.str.removeprefix("tags.")

    return df_jobs


def qoi_gp_deterministic(
    env_dataset: Dataset[NDArray[np.float64]],
    n_periods: int,
    n_posterior_samples: int,
    n_qoi_runs: int,
    jobs_output_file: None | Path = None,
):
    """Run the QoI with a highly trained, deterministic model.

    Args:
        env_dataset: Dataset containing the envdata we have access to.
        n_periods (int): number of periods to use. each produces `n_qoi_runs` estimates of the ERD.
        n_posterior_samples (int): number of posterior samples to take. Each produces an estimate of the QoI.
        n_qoi_runs (int): Number of times to run the QoIEstimator (a single run produces n_posterior_sampler estimates)
        jobs_output_file (None | Path): Json file to write raw run results to. If not provided do no write results.

    Return:
        A DataFrame where each row represents a single QoIeEstimator run. Contains columns:
            - mean: The samples mean produced by the QoIeEstimator for that run.
            - var: The samples var produced by the QoIeEstimator for that run.
            - samples: The samples produced by the QoIeEstimator for that run.
            - name: Name of the estimator. This is used to group these results together.

    Details:
        We want to check the GP is a good fit, because if not the results will be bias regardless of other settings.
        - we start by checking if the posterior mean is a good fit.

    Expected results:
        We would expect this to produce a very similar result to the `qoi_no_gp`.
        - produces the same results as the ground truth.
        - May have slight bias due to slight model misfit.
    """
    botorch_model = get_trained_gp()

    gp_deterministic = PosteriorMeanModel(botorch_model)

    dataloader = get_dataloader(env_dataset, n_periods, batch_size=64)
    qoi_gp_deterministic_jobs = []
    for i in range(n_qoi_runs):
        qoi_est = GPBruteForce(
            # random dataloader give different env samples for each instance
            env_iterable=dataloader,
            # IndexSampler needs to be used with GenericDeterministicModel. Each sample just selects the mean.
            posterior_sampler=IndexSampler(torch.Size([n_posterior_samples])),
            no_grad=True,
        )

        qoi_gp_deterministic_jobs.append(
            QoIJob(
                name=f"qoi_gp_deterministic_{i}",
                qoi=qoi_est,
                model=gp_deterministic,
                tags={"name": "qoi_gp_deterministic"},
            )
        )

    qoi_gp_deterministic_results = [job(output_file=jobs_output_file) for job in tqdm(qoi_gp_deterministic_jobs)]
    df_jobs = pd.json_normalize([item.to_dict() for item in qoi_gp_deterministic_results], max_level=1)
    df_jobs.columns = df_jobs.columns.str.removeprefix("tags.")

    return df_jobs


def get_trained_gp(n_points: int = 512) -> SingleTaskGP:
    """Create a botorch model."""
    # TODO(sw): would be good to pull this out of here
    exp = make_experiment()
    add_sobol_points_to_experiment(exp, n_iter=n_points, seed=7)

    botorch_model_bridge = Models.BOTORCH_MODULAR(
        experiment=exp,
        data=exp.fetch_data(),
    )

    input_transform, outcome_transform = transforms.ax_to_botorch_transform_input_output(
        transforms=list(botorch_model_bridge.transforms.values()), outcome_names=botorch_model_bridge.outcomes
    )
    botorch_model = botorch_model_bridge.model.surrogate.model
    botorch_model.outcome_transform = outcome_transform
    botorch_model.input_transform = input_transform

    return botorch_model


def qoi_no_gp(
    env_dataset: Dataset[NDArray[np.float64]],
    n_periods: int,
    n_posterior_samples: int,
    n_qoi_runs: int,
    jobs_output_file: None | Path = None,
) -> pd.DataFrame:
    """Test the QoIEstimator with the true underlying function.

    Args:
        env_dataset: Dataset containing the envdata we have access to.
        n_periods (int): number of periods to use. each produces `n_qoi_runs` estimates of the ERD.
        n_posterior_samples (int): number of posterior samples to take. Each produces an estimate of the QoI.
        n_qoi_runs (int): Number of times to run the QoIEstimator (a single run produces n_posterior_sampler estimates)
        jobs_output_file (None | Path): Json file to write raw run results to. If not provided do no write results.

    Return:
        A DataFrame where each row represents a single QoIeEstimator run. Contains columns:
            - mean: The samples mean produced by the QoIeEstimator for that run.
            - var: The samples var produced by the QoIeEstimator for that run.
            - samples: The samples produced by the QoIeEstimator for that run.
            - name: Name of the estimator. This is used to group these results together.

    Details:
    When using the true underlying (rather than a GP) there are still difference between the QoIEstimator and the
    Ground truth. This is caused by the difference in how env data is used:
    - Ground truth: each period gets a unique sample of env data
        - (e.g there are n_erd_samples * n_ests unique periods of env data)
    - QoIEstimator: All estimates produce by a single run see the same set of periods
        - (e.g there are n_erd_samples unique periods of env data)

    If insuffeciently large n_periods is used:
    - best guess will have more noise, (individual samples will be more biased), the overall there should remain no bias
    - the bounds on the distribution will be smaller than the real value

    This checks that this value is approapriately set. If not, the other QoIEstimator (e.g using Gps etc) will not
    perform.

    Expected result:
    - Produces the same results as the ground truth.
    """

    def true_scale_func_tensor(x: torch.Tensor) -> torch.Tensor:
        return torch.from_numpy(simulator._true_scale_func(x.numpy()))

    def true_loc_func_tensor(x: torch.Tensor) -> torch.Tensor:
        return torch.from_numpy(simulator._true_loc_func(x.numpy()))

    def true_underling_func(x: torch.Tensor) -> torch.Tensor:
        """Put true underling function into format required for GenericDeterministicModel.

        Args:
        x: tensor of dimensions (*b, n, d)

        returns:
        tensor of dimensions (*b, n, d)
        """
        locs = true_loc_func_tensor(x).unsqueeze(-1)
        scales = true_scale_func_tensor(x).unsqueeze(-1)
        return torch.concat([locs, scales], dim=-1)

    # This is roughly equivalent to useing a GP the no variance
    # Running this model produces  botorch.posteriors.ensemble.EnsemblePosterior. They require index samplers
    det_model = GenericDeterministicModel(true_underling_func, num_outputs=2)
    dataloader = get_dataloader(env_dataset, n_periods)
    qoi_no_gp_jobs = []
    for i in range(n_qoi_runs):
        qoi_est = GPBruteForce(
            # random dataloader give different env samples for each instance
            env_iterable=dataloader,
            # IndexSampler needs to be used with GenericDeterministicModel. Each sample just selects the mean.
            posterior_sampler=IndexSampler(torch.Size([n_posterior_samples])),
            no_grad=True,
        )

        qoi_no_gp_jobs.append(QoIJob(name=f"qoi_no_gp_{i}", qoi=qoi_est, model=det_model, tags={"name": "qoi_no_gp"}))

    qoi_no_gp_results = [job(output_file=jobs_output_file) for job in tqdm(qoi_no_gp_jobs)]

    df_jobs = pd.json_normalize([item.to_dict() for item in qoi_no_gp_results], max_level=1)
    df_jobs.columns = df_jobs.columns.str.removeprefix("tags.")
    return df_jobs


def get_dataloader(dataset: Dataset[NDArray[np.float64]], n_periods: int, batch_size: int = 64):
    """Helper to create a Dataloader which randomly samples from the dataset.

    Dataloader samples with replacement from the dataset. Total iterable will have shape
    (n_periods, N_ENV_SAMPLES_PER_PERIOD). If the dataloader is reused, it will generate new random samples.

    Args:
        dataset: The available data to draw samples from
        n_periods: Number of periods of data to produce.
        batch_size: N_ENV_SAMPLES_PER_PERIOD will be split into batches of this size
    """
    generator = torch.Generator()
    _ = generator.manual_seed(7)
    replacement_sampler = RandomSampler(
        data_source=dataset,  # type: ignore[arg-type]
        num_samples=n_periods * N_ENV_SAMPLES_PER_PERIOD,
        generator=generator,
        replacement=True,
    )
    batch_sampler_rand = BatchInvariantSampler2d(
        sampler=replacement_sampler, batch_shape=torch.Size([n_periods, batch_size])
    )
    return DataLoader(dataset, batch_sampler=batch_sampler_rand)


def ground_truth_estimate(
    n_periods: int, n_posterior_samples: int, n_qoi_runs: int, jobs_output_file: None | Path = None
) -> pd.DataFrame:
    """For a given budget of ERD samples, what answer is achieved using the true undelying directly (perfect knowledge).

    If other QoI outputs have:
    - More noise: likely need to increase the samples of different source of randomness (e.g gp)
    - Less noise: likely have frozen/subsampled sources of randomness

    The behaviour of the mean and variance of each distribution can help analyse these issues.

    Args:
        n_periods (int): number of periods to use. each produces `n_qoi_runs` estimates of the ERD.
        n_posterior_samples (int): number of posterior samples to take. Each produces an estimate of the QoI.
        n_qoi_runs (int): Number of times to run the QoIEstimator (a single run produces n_posterior_sampler estimates)
        jobs_output_file (None | Path): Json file to write raw run results to. If not provided do no write results.

    For each sample from the GP (e.g a potential underling):
        - Number of ERD samples = n_periods * n_qoi_runs

    Return:
        A DataFrame where each row represents a single QoIeEstimator run. Contains columns:
            - mean: The samples mean produced by the QoIeEstimator for that run.
            - var: The samples var produced by the QoIeEstimator for that run.
            - samples: The samples produced by the QoIeEstimator for that run.
            - name: Name of the estimator. This is used to group these results together.
    """
    brute_force_erd_samples = brute_force.collect_or_calculate_results(
        period_length=N_ENV_SAMPLES_PER_PERIOD, num_estimates=300_000
    )

    # Turn this into job results to we can plot in the same manner as above
    ground_truth_results = []
    for _ in range(n_qoi_runs):
        qoi_comparable_output = qoi_comparable_output_using_full_brute_force(
            n_erd_samples=n_periods, n_ests=n_posterior_samples, brute_force_erd_samples=brute_force_erd_samples
        )
        # Directly skip to results here as nothing is required to run.
        ground_truth_results.append(
            QoIJobResult(
                mean=qoi_comparable_output.mean(),
                var=qoi_comparable_output.var(),
                samples=qoi_comparable_output,
                tags={"name": "ground_truth"},
            )
        )

    if jobs_output_file:
        for result in ground_truth_results:
            utils.append_to_json(result.to_dict(), jobs_output_file)

    df_jobs = pd.json_normalize([item.to_dict() for item in ground_truth_results], max_level=1)
    df_jobs.columns = df_jobs.columns.str.removeprefix("tags.")
    return df_jobs


def qoi_comparable_output_using_full_brute_force(
    n_erd_samples: int, n_ests: int, brute_force_erd_samples: torch.Tensor
) -> torch.Tensor:
    """Mimics the output the QoIEstimator produces, but uses precalculated ERD samples directly.

    Purpose is to show what output would expect from a 'perfect' QoIEstimator with this number of samples.
        - 'perfect' Meaning no GP uncertainty, and no subsampling/freezing. Achieved because precalc results:
            - use the true underling function
            - use a unique sample of weather for each ERD sample/

    Args:
        n_erd_samples: the number of erd samples used to make a single estimate of the QoI.
        - Note: in QoIestimate the number of ERD samples is controlled by n_periods.
        n_ests: The number of QoI estimates to make.
        - Note: in QoIestimate the number of estimates made is controlle by n_posterior_samples
        brute_force_erd_samples: (n,) precalculated ERD samples. Usecases often contain methods for bruteforce.
            - Should be at least: n > n_erd_samples * n_ests
            - ideally n > n_erd_samples * n_ests * n_qoi_runs

    Return:
        `n_ests` for the QoI. This output is comparable with the output returned by the QoIEstimator interface.
    """
    sample_idx = torch.randint(0, len(brute_force_erd_samples), size=torch.Size([n_ests, n_erd_samples]))
    samples = brute_force_erd_samples[sample_idx]
    return samples.median(dim=-1).values  # noqa: PD011


# %%
if __name__ == "__main__":
    # %% For manual inspection/running of the system
    test_qoi_brute_force_system_test(output_dir=None, n_qoi_runs=120, run_tests=True, show_plots=True)  # type: ignore  # noqa: PGH003

    # %%
    """This following is a helper to determine the bounds testing in qoi_estimator_output.

    Here we first run the test with a large number of samples, then with subsampling estimate what the bounds would be
    if running with fewer samples. Now that the bounds are calibrated, we can set the pytest runs to use
    fewer samples, which takes less time.

    The following is the process that can be used to get new calibrated bounds (e.g if the estimators or number of
    subsamples changes).
    """
    import matplotlib.pyplot as plt

    _ = plt.ioff()  # stop plots from displaying for repeat runs
    # %% Args to set
    total_number_of_runs = 400
    subsample_size = 50

    # %%
    # Make a large number of samples once off:
    output_dir_path = Path("results") / get_id()
    output_dir_path.mkdir(exist_ok=True)

    _ = test_qoi_brute_force_system_test(
        output_dir=output_dir_path,
        n_qoi_runs=total_number_of_runs,  # make as large as feasible.
        run_tests=False,
    )

    # %%
    # read the Qoi_jons
    with (output_dir_path / "qoi_job_results.json").open("r") as fp:
        qoi_job_results = json.load(fp)

    full_df = pd.json_normalize(qoi_job_results, max_level=1)
    full_df.columns = full_df.columns.str.removeprefix("tags.")

    # %% Subsample the results
    def sample_group(group, n=subsample_size):  # type: ignore  # noqa: ANN001, PGH003
        return group.sample(n=min(len(group), n))

    # Group by 'name' and sample 50 items from each group
    all_statistics = []
    for _ in range(100):
        sampled_df = full_df.groupby("name", group_keys=False).apply(sample_group)
        statistics = test_qoi_brute_force_system_test(
            output_dir=None, jobs_input_file=sampled_df, run_tests=False, show_plots=False, return_statistics=True
        )
        all_statistics.append(statistics)

    df = pd.json_normalize(all_statistics, max_level=1)  # type: ignore  # noqa: PD901, PGH003
    df.head()  # type: ignore  # noqa: PGH003

    # %%
    # fmt: off
    """For each group, calculate the same statistics that we need to test"""

    # Ground truth
    print("best_guess_z: ",df["ground_truth.best_guess_z"].abs().max())  # noqa: T201

    # %%
    #  qoi_no_gp
    print("best_guess_z: ",df["qoi_no_gp.best_guess_z"].abs().max()) # noqa: T201

    print("best_guess_std\n",(df["qoi_no_gp.best_guess_std"] / df["ground_truth.best_guess_std"]).agg(["min","max"])) # noqa: T201
    print("var_mean\n",(df["qoi_no_gp.var_mean"] / df["ground_truth.var_mean"]).agg(["min","max"])) # noqa: T201
    print("var_std\n",(df["qoi_no_gp.var_std"] / df["ground_truth.var_std"]).agg(["min","max"])) # noqa: T201

    # %% qoi_gp_deterministic
    print("best_guess_z: ", df["qoi_gp_deterministic.best_guess_z"].abs().max()) # noqa: T201

    # %% qoi_gp_low_uncertainty
    print("best_guess_z: ", df["qoi_gp_low_uncertainty.best_guess_z"].abs().max()) # noqa: T201

    print("best_guess_std\n",(df["qoi_gp_low_uncertainty.best_guess_std"] / df["ground_truth.best_guess_std"]).agg(["min","max"]))  # noqa: E501, T201
    print("var_mean\n",(df["qoi_gp_low_uncertainty.var_mean"] / df["ground_truth.var_mean"]).agg(["min","max"])) # noqa: T201
    print("var_std\n",(df["qoi_gp_low_uncertainty.var_std"] / df["ground_truth.var_std"]).agg(["min","max"])) # noqa: T201

    # fmt:on
    # %%



"""Dataset that return importance sample information."""
# pyright: reportUnnecessaryTypeIgnoreComment=false

# @TODO: The Importance...Wrapper classes break compatibility with their base class,
#        as they return a tuple instead of the expected data type.
#        On thing is that it feels a bit "dirty" from an architectural point of view.
#        More important would be if the implementation causes problems when using
#        the Importance...Wrapper Dataset classes with PyTorch's DataLoader,
#        as I suspect this expects `__get_item__()` to retrun not a tuple, but a single object (of type `T_co`).
#        To suppress pyright errors, I for now deactivated the corresponding error codes for this module (see below).
#        However, maybe there is a cleaner implementation to add weights to a dataset? @swinter @KristofferSkare
#        (_CoPilot suggestion_, not checked):
#        "One idea would be to use a `collate_fn` in the DataLoader, which could add the weights to the data."
#        @ClaasRostock, 2024-08-30
# pyright: reportIndexIssue=false
# pyright: reportReturnType=false
# pyright: reportIncompatibleMethodOverride=false
# mypy: disable-error-code="override, index, return-value"

import numpy as np
import torch
from torch.utils.data import Dataset
from torch.utils.data.dataset import T_co


class ImportanceIndexWrapper(Dataset[T_co]):
    """Wraps an existing dataset, returning part of the item as the importance weight."""

    def __init__(self, dataset: Dataset[T_co], importance_idx: int) -> None:
        """Wrap an existing dataset, returning part of the item as the importance weight.

        Args:
            dataset: A dataset, where one of the columns should be used as an importance weight.
                - Note: Currently only datasets that return numpy or torch tensors are supported
            importance_idx: the column index containing the importance weights.

        Todo:
            - Generalise this to deal with other types of dataset output (list, numpy etc).
        """
        self.dataset: Dataset[T_co] = dataset
        self.importance_idx: int = importance_idx

        datapoint = self.dataset[0]

        # Might be more pythonic to not check up front, but I think its a better user experience
        # These have the same indexing interface so are easy to support
        if not (isinstance(datapoint, np.ndarray | torch.Tensor)):
            msg = (
                "ImportanceIndexWrapper only support datasets where the underlying data is"
                f"torch.Tensor or NDArray[np.float64]. Instead got {type(datapoint)}"
            )
            raise TypeError(msg)

        self.mask = torch.ones(datapoint.shape, dtype=torch.bool)
        self.mask[importance_idx] = False

    def __len__(self) -> int:  # noqa: D105
        return len(self.dataset)  # type: ignore[arg-type]

    def __getitem__(self, index: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Gets the data and importance weight at the requested index in the dataset.

        Args:
            index: what index in the dataset should be returned

        Returns:
            A tuple where:
            - Element 0: (tensor) the datapoint
            - Element 1: Float representing the importance weight
        """
        datapoint = self.dataset[index]
        # datapoint with importance column, importance weight
        return datapoint[self.mask], datapoint[self.importance_idx]


class ImportanceAddedWrapper(Dataset[T_co]):
    """Combine existing dataset with one containing the related importance weights."""

    def __init__(self, data_dataset: Dataset[T_co], importance_dataset: Dataset[T_co]) -> None:
        """Combines one Dataset of data, with a Dataset of importance weights.

        Args:
            data_dataset: Dataset containing input data
            importance_dataset: Contains importance weights.
                - `__get_item__()` should return floats.

        Note:
            Both datasets need to share the same index.
            E.g. `importance_datasets[idx]` provides the importance weight for the data `data_dataset[idx]`.
        """
        if len(data_dataset) != len(importance_dataset):  # type: ignore[arg-type]
            msg = (
                "Expected data_dataset and importance_dataset to have sample length."
                f" Instead got {len(data_dataset)=}, and {len(importance_dataset)}."  # type: ignore[arg-type]
            )
            raise ValueError(msg)

        self.data_dataset = data_dataset
        self.importance_dataset = importance_dataset

    def __len__(self) -> int:  # noqa: D105
        return len(self.data_dataset)  # type: ignore[arg-type]

    def __getitem__(self, index: int) -> tuple[torch.Tensor, torch.Tensor]:
        """Gets the data and importance weight at the requested index in the dataset."""
        # datapoint with importance column, importance weight
        return self.data_dataset[index], self.importance_dataset[index]
